from typing import List, Optional, Union

import pydantic


class HamiltonianReduction(pydantic.BaseModel):
    two_qubit_reduction: Optional[bool] = pydantic.Field(default=False)
    number_particle_reduction: Optional[bool] = pydantic.Field(default=False)
    freeze_core: Optional[bool] = pydantic.Field(default=False)
    remove_orbitals: Optional[List[int]] = pydantic.Field(
        default=None, description="list of orbitals to remove"
    )
    z2_symmetries: Optional[Union[str, List[int]]] = pydantic.Field(
        default=None,
        description="possible values are: None, 'auto' or a list of 1 and -1",
    )



import numpy as np
import pyomo.core as pyo


def portfolio_optimization(
    covariances: np.ndarray, returns: np.ndarray, budget: int
) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    num_assets = len(returns)
    model.x = pyo.Var(range(num_assets), domain=pyo.Binary)
    x_array = model.x.values()

    model.budget = pyo.Constraint(expr=(sum(x_array) == budget))

    risk: float = x_array @ covariances @ x_array
    profit: float = returns @ x_array
    model.risk, model.profit = risk, profit

    model.cost = pyo.Objective(expr=model.risk - model.profit, sense=pyo.minimize)

    return model



from enum import Enum
from typing import Dict, Optional

import pydantic

from classiq_interface.generator.arith.arithmetic import (
    DEFAULT_ARG_NAME,
    DEFAULT_OUT_NAME,
)
from classiq_interface.generator.arith.register_user_input import RegisterUserInput
from classiq_interface.generator.function_params import FunctionParams


class UnaryOpParams(FunctionParams):
    arg: RegisterUserInput
    output_size: Optional[pydantic.PositiveInt]
    output_name: Optional[str]
    inplace: bool = False

    def create_io_enums(self):
        output_name = self.output_name if self.output_name else DEFAULT_OUT_NAME
        arg_name = self.arg.name if self.arg.name else DEFAULT_ARG_NAME

        output_name_dict: Dict[str, str] = {output_name: output_name}
        input_name_dict = {arg_name: arg_name}
        if not self.inplace:
            output_name_dict.update(input_name_dict)

        self._output_enum = Enum("BinaryOpOutputs", output_name_dict)
        self._input_enum = Enum("UnaryOpInputs", input_name_dict)

    class Config:
        arbitrary_types_allowed = True


class BitwiseInvert(UnaryOpParams):
    pass


class Negation(UnaryOpParams):
    pass



import math

import pydantic

from classiq_interface.generator.arith.register_user_input import RegisterUserInput

MAX_FRACTION_PLACES = 5


class FixPointNumber(pydantic.BaseModel):
    float_value: float
    max_fraction_places: int = MAX_FRACTION_PLACES
    _is_signed: bool = pydantic.PrivateAttr(default=False)
    _fraction_places: int = pydantic.PrivateAttr(default=0)
    _int_val: int = pydantic.PrivateAttr(default=0)
    _size: pydantic.PositiveInt = pydantic.PrivateAttr(default=1)
    _integer_part_size: pydantic.PositiveInt = pydantic.PrivateAttr(default=1)
    _is_initiated: bool = pydantic.PrivateAttr(default=False)

    def _initiate_values(self) -> None:
        if self.float_value < 0.0:
            self._is_signed = True
        self.set_int_representation()
        self._size = self.bit_length(should_ensure_initiation=False)
        self._integer_part_size = self._size - self._fraction_places
        self._is_initiated = True

    def set_int_representation(self) -> None:
        int_val = int(self.float_value * 2 ** self.max_fraction_places)
        int_val = self.signed_int_to_unsigned(int_val)

        if int_val == 0:
            fraction_places = 0
        else:
            bin_val = bin(int_val)[2:]
            fraction_places = self.max_fraction_places
            for b in reversed(bin_val):
                if b == "1" or fraction_places == 0:
                    break
                fraction_places -= 1
            int_val = int_val >> (self.max_fraction_places - fraction_places)

        self._fraction_places = fraction_places
        self._int_val = int_val

    @staticmethod
    def signed_int_to_unsigned(number: int):
        """Return the integer value of a signed int if it would we read as un-signed in binary representation"""
        if number >= 0:
            return number

        not_power2 = abs(number) & (abs(number) - 1) != 0
        return number + 2 ** (number.bit_length() + 1 * not_power2)

    @staticmethod
    def binary_to_float(bin_rep: str, fraction_part_size: int, is_signed: bool = False):
        negative_offset = -(2 ** len(bin_rep)) * (bin_rep[0] == "1") * is_signed
        value = int(bin_rep, 2) + negative_offset
        if (
            fraction_part_size > 0
        ):  # separated the clause to so that the value remains int if there is no fraction part
            value = value / 2 ** fraction_part_size
        return value

    @property
    def is_signed(self) -> bool:
        if not self._is_initiated:
            self._initiate_values()
        return self._is_signed

    @property
    def fraction_places(self):
        if not self._is_initiated:
            self._initiate_values()
        return self._fraction_places

    def set_fraction_places(self, value: int):
        if value < self._fraction_places:
            raise ValueError("size cannot be lower than minimum number bits required")

        if value > self.max_fraction_places:
            self.max_fraction_places = value
            self._initiate_values()

        self._int_val = math.floor(self.int_val * 2 ** (value - self.fraction_places))
        self._fraction_places = value
        self._size = self._integer_part_size + self._fraction_places

    @property
    def int_val(self):
        if not self._is_initiated:
            self._initiate_values()
        return self._int_val

    @property
    def integer_part_size(self):
        if not self._is_initiated:
            self._initiate_values()
        return self._integer_part_size

    def set_integer_part_size(self, value: int):
        if value < self.integer_part_size:
            raise ValueError("size cannot be lower than minimum number bits required")
        self._integer_part_size = value
        self._size = self._integer_part_size + self._fraction_places
        self._int_val = int(self.bin_val, 2)

    def bit_length(self, *, should_ensure_initiation=True) -> pydantic.PositiveInt:
        if should_ensure_initiation and not self._is_initiated:
            self._initiate_values()
        return 1 if self._int_val == 0 else self._int_val.bit_length()

    @property
    def size(self):
        if not self._is_initiated:
            self._initiate_values()
        return self._size

    def __len__(self):
        return self.size

    @property
    def bin_val(self):
        if not self._is_initiated:
            self._initiate_values()

        bin_rep = bin(self._int_val)[2:]
        size_diff = self.size - len(bin_rep)
        if self.float_value >= 0:
            return "0" * size_diff + bin_rep
        else:
            return "1" * size_diff + bin_rep

    @property
    def actual_float_value(self):
        return self.binary_to_float(self.bin_val, self.fraction_places, self._is_signed)

    def to_register_user_input(self) -> RegisterUserInput:
        return RegisterUserInput(
            size=self.size,
            is_signed=self.is_signed,
            fraction_places=self.fraction_places,
        )

    def __eq__(self, other):
        return self.actual_float_value == other

    def __ge__(self, other):
        return self.actual_float_value >= other

    def __gt__(self, other):
        return self.actual_float_value > other

    def __le__(self, other):
        return self.actual_float_value <= other

    def __lt__(self, other):
        return self.actual_float_value < other

    def __ne__(self, other):
        return self.actual_float_value != other

    def __getitem__(self, item):
        return [v for v in self.bin_val[::-1]][
            item
        ]  # follow qiskit convention that LSB is the top wire, bigendian

    def __neg__(self):
        return FixPointNumber(
            float_value=-self.float_value, max_fraction_places=self.max_fraction_places
        )

    class Config:
        extra = "forbid"



from typing import List, Optional

import pydantic

from classiq_interface.generator import function_params

LENGTH_ERROR_MESSAGE = "Input length error: Length of weights vector has to be identical to the number of input qubits"
STATE = "state"
SUM = "sum"


class pydanticStrictNonNegativeInteger(pydantic.ConstrainedInt):
    strict = True
    ge = 0


class WeightedAdder(function_params.FunctionParams):
    """
    Creates a circuit implementing a scalar product between an n-qubit state |q1,q2,...,qn> and an n-length non-negative
    integer vector (w1,w2,...wn), such that the result of the output register is |q1*w1+q2*w2+...qn*wn>.
    If no weights are provided, they are default to 1 for every qubit.
    """

    num_state_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of input qubits"
    )
    weights: Optional[List[pydanticStrictNonNegativeInteger]] = pydantic.Field(
        description="List of non-negative integers corresponding to the weight of each qubit"
    )
    inverse: bool = pydantic.Field(
        default=False, description="Create the inverse gate of weighted adder"
    )

    @pydantic.validator("weights", always=True)
    def validate_weights(cls, weights, values):
        num_state_qubits = values.get("num_state_qubits")
        if weights is None:
            return [1] * num_state_qubits
        if len(weights) != num_state_qubits:
            raise ValueError(LENGTH_ERROR_MESSAGE)
        return weights

    def _create_io_names(self):
        if self.inverse:
            self._input_names = [STATE, SUM]
        else:
            self._input_names = [STATE]
        self._output_names = [STATE, SUM]



from typing import Generic, Optional, TypeVar

import pydantic
from pydantic.generics import GenericModel

RangeType = TypeVar("RangeType", int, float)


class Range(GenericModel, Generic[RangeType]):
    lower_bound: Optional[RangeType] = None
    upper_bound: Optional[RangeType] = None

    @pydantic.validator("upper_bound")
    def validate_bounds_order(cls, upper_bound, values):
        lower_bound = values.get("lower_bound")

        if (
            lower_bound is not None
            and upper_bound is not None
            and lower_bound > upper_bound
        ):
            raise ValueError("lower bound must not be greater than upper bound")

        return upper_bound


NonNegativeIntRange = Range[pydantic.NonNegativeInt]
NonNegativeFloatRange = Range[pydantic.NonNegativeFloat]



import networkx as nx
import pyomo.core as pyo


def arithmetic_eq(x1: int, x2: int) -> int:
    return x1 * x2 + (1 - x1) * (1 - x2)


def maxcut(graph: nx.Graph) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    model.x = pyo.Var(graph.nodes, domain=pyo.Binary)

    model.cost = pyo.Objective(
        expr=sum(
            arithmetic_eq(model.x[node1], model.x[node2])
            for (node1, node2) in graph.edges
        ),
        sense=pyo.minimize,
    )

    return model



from typing import Optional

import pydantic

from classiq_interface.generator import function_params


class StandardGate(function_params.FunctionParams):
    _num_target_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=1)

    _num_ctrl_qubits: Optional[pydantic.PositiveInt] = pydantic.PrivateAttr(
        default=None
    )

    _input_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_INPUT_NAME])
    _output_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_OUTPUT_NAME])

    @property
    def num_ctrl_qubit(self):
        return self._num_ctrl_qubits

    @property
    def num_target_qubits(self):
        return self._num_target_qubits


class XGate(StandardGate):
    """
    creates a X gate
    """

    pass


class YGate(StandardGate):
    """
    creates a Y gate
    """

    pass


class ZGate(StandardGate):
    """
    create a Z gate
    """

    pass


class HGate(StandardGate):
    """
    creates a Hadamard gate
    """

    pass


class IGate(StandardGate):
    """
    creates the identity gate
    """

    pass


class SGate(StandardGate):
    """
    Z**0.5
    """

    pass


class SdgGate(StandardGate):
    """
    creates the inverse S gate
    """

    pass


class SXGate(StandardGate):
    """
    X**0.5
    """

    pass


class SXdgGate(StandardGate):
    """
    creates the inverse SX gate
    """

    pass


class TGate(StandardGate):
    """
    Z**0.25
    """

    pass


class TdgGate(StandardGate):
    """
    creates the inverse T gate
    """

    pass


class SwapGate(StandardGate):
    _num_target_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=2)
    """
    Swaps between two qubit states
    """


class iSwapGate(StandardGate):
    _num_target_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=2)
    """
    Swaps between two qubit states and add phase of i to the amplitudes of |01> and |10>
    """



from typing import Dict, Optional, Tuple, Type

import pydantic

from classiq_interface.generator.functions import function_data
from classiq_interface.generator.user_defined_function_params import CustomFunction
from classiq_interface.helpers.custom_pydantic_types import pydanticFunctionNameStr

DEFAULT_FUNCTION_LIBRARY_NAME = "default_function_library_name"


class FunctionLibraryData(pydantic.BaseModel):
    """Facility to store user-defined custom functions."""

    name: pydanticFunctionNameStr = pydantic.Field(
        default=DEFAULT_FUNCTION_LIBRARY_NAME,
        description="The name of the custom function library",
    )

    function_dict: Dict[str, function_data.FunctionData] = pydantic.Field(
        default_factory=dict,
        description="A dictionary containing the custom functions in the library.",
    )

    functions: Tuple[function_data.FunctionData, ...] = pydantic.Field(
        default=tuple(),
        description="A tuple for inputting custom functions to the library.",
    )

    @pydantic.validator("name")
    def validate_name(cls, name: str):
        function_data.validate_name_end_not_newline(name=name)
        return name

    @pydantic.validator("function_dict")
    def validate_function_dict(
        cls, function_dict: Dict[str, function_data.FunctionData]
    ):
        if not all(
            function_data.name == name for name, function_data in function_dict.items()
        ):
            raise AssertionError("Bad function_dict encountered.")
        return function_dict

    @pydantic.validator("functions")
    def validate_functions(
        cls, functions: Optional[Tuple[function_data.FunctionData, ...]], values
    ):
        if not functions:
            return None
        if values.get("function_dict"):
            raise ValueError("Expected only a single function data input field")
        values["function_dict"] = {
            function_data.name: function_data for function_data in functions
        }
        return None

    @staticmethod
    def validate_function_in_library(
        function_params: CustomFunction,
        library: Optional["FunctionLibraryData"],
        error_handler: Type[Exception] = ValueError,
    ) -> None:
        if library is None or function_params.name not in library.function_dict:
            raise error_handler("The function is not found in included library.")



from enum import Enum

import pydantic

from classiq_interface.chemistry import hamiltonian_reduction, molecule


class FermionMapping(str, Enum):
    JORDAN_WIGNER = "jordan_wigner"
    PARITY = "parity"
    BRAVYI_KITAEV = "bravyi_kitaev"
    FAST_BRAVYI_KITAEV = "fast_bravyi_kitaev"


class GroundStateProblem(pydantic.BaseModel):
    molecule: molecule.Molecule
    basis: str = pydantic.Field(default="sto3g", description="Molecular basis set")
    mapping: FermionMapping = pydantic.Field(
        default=FermionMapping.JORDAN_WIGNER, description="Fermionic mapping type"
    )
    reductions: hamiltonian_reduction.HamiltonianReduction



from typing import List

import numpy as np
import pyomo.core as pyo


def integer_portfolio_optimization(
    covariances: np.ndarray, returns: np.ndarray, upper_bounds: List[int]
) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    num_assets = len(returns)

    def bounds(_, i):
        return 0, upper_bounds[i]

    model.x = pyo.Var(range(num_assets), domain=pyo.NonNegativeIntegers, bounds=bounds)

    x_array: np.ndarray = np.array(model.x.values())
    risk: float = x_array @ covariances @ x_array
    profit: float = returns @ x_array
    model.risk, model.profit = risk, profit

    model.cost = pyo.Objective(expr=model.risk - model.profit, sense=pyo.minimize)

    return model



import base64
import enum
import io
import json
import tempfile
import webbrowser
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

import pydantic
from PIL import Image

from classiq_interface.generator.generation_metadata import GenerationMetadata
from classiq_interface.generator.synthesis_metrics import SynthesisMetrics

_MAXIMUM_STRING_LENGTH = 250


class LongStr(str):
    def __repr__(self):
        if len(self) > _MAXIMUM_STRING_LENGTH:
            length = len(self)
            return f'"{self[:4]}...{self[-4:]}" ({length=})'
        return super().__repr__()


class QuantumFormat(str, enum.Enum):
    QASM = "qasm"
    QSHARP = "qs"
    QIR = "ll"
    IONQ = "ionq"
    CIRQ_JSON = "cirq_json"
    QASM_CIRQ_COMPATIBLE = "qasm_cirq_compatible"


class GenerationStatus(str, enum.Enum):
    NONE = "none"
    SUCCESS = "success"
    UNSAT = "unsat"
    TIMEOUT = "timeout"
    CANCELLED = "cancelled"
    ERROR = "error"


# TODO: Merge all output formats to single field (dictionary?) to avoid clutter
class GeneratedCircuit(pydantic.BaseModel):
    qubit_count: Optional[int]
    depth: Optional[int]
    qasm: Optional[str]
    transpiled_qasm: Optional[str]
    qsharp: Optional[str]
    qir: Optional[str]
    ionq: Optional[str]
    cirq_json: Optional[str]
    qasm_cirq_compatible: Optional[str]
    output_format: List[QuantumFormat]
    image_raw: str
    interactive_html: Optional[str]
    metadata: Optional[GenerationMetadata]
    synthesis_metrics: Optional[SynthesisMetrics]

    def show(self) -> None:
        self.image.show()

    def show_interactive(self):
        with tempfile.NamedTemporaryFile(
            "w", delete=False, suffix="_interactive_circuit.html"
        ) as f:
            url = f"file://{f.name}"
            f.write(self.interactive_html)
        webbrowser.open(url)

    @property
    def image(self):
        return Image.open(io.BytesIO(base64.b64decode(self.image_raw)))

    def save_results(self, filename: Optional[Union[str, Path]] = None) -> None:
        """
        Saves generated circuit results as json.
            Parameters:
                filename (Union[str, Path]): Optional, path + filename of file.
                                             If filename supplied add `.json` suffix.

            Returns:
                  None
        """
        if filename:
            path = Path(filename)
        else:
            path = Path(
                f"generated_circuit_{datetime.now().strftime('%m_%d_%Y_%H_%M_%S')}.json"
            )
        path.write_text(json.dumps(self.dict(), indent=4))

    @pydantic.validator("qasm", "transpiled_qasm", "image_raw", "interactive_html")
    def reformat_long_strings(cls, v):
        if v is None:
            return v
        return LongStr(v)


class GenerationResult(pydantic.BaseModel):
    status: GenerationStatus
    details: Union[GeneratedCircuit, str]



import pydantic

from classiq_interface.combinatorial_optimization.solver_types import QSolver


class QAOAPreferences(pydantic.BaseModel):
    qsolver: QSolver = pydantic.Field(
        default=QSolver.QAOAPenalty,
        description="Indicates whether to use QAOA with penalty terms or constrained QAOA.",
    )
    qaoa_reps: pydantic.PositiveInt = pydantic.Field(
        default=1, description="Number of layers in qaoa ansatz."
    )
    penalty_energy: float = pydantic.Field(
        default=None,
        description="Penalty energy for invalid solutions. The value affects "
        "the converges rate. Small positive values are preferred",
    )

    @pydantic.validator("penalty_energy", pre=True, always=True)
    def check_penalty_energy(cls, penalty_energy, values):
        qsolver = values.get("qsolver")
        if penalty_energy is not None and qsolver != QSolver.QAOAPenalty:
            raise ValueError("Use penalty_energy only for QSolver.QAOAPenalty.")

        if penalty_energy is None and qsolver == QSolver.QAOAPenalty:
            penalty_energy = 2

        return penalty_energy



import pydantic

from classiq_interface.generator.standard_gates.standard_gates import StandardGate


class RXGate(StandardGate):
    """
    Rotation by theta about the X axis
    """

    theta: float


class RYGate(StandardGate):
    """
    Rotation by theta about the Y axis
    """

    theta: float


class RZGate(StandardGate):
    """
    Rotation by phi about the Z axis
    """

    phi: float


class RGate(StandardGate):
    """
    Rotation by theta about the cos(phi)X + sin(phi)Y axis
    """

    theta: float
    phi: float


class PhaseGate(StandardGate):
    """
    Add relative phase of lambda
    """

    theta: float


class DoubleRotationGate(StandardGate):
    """
    Base class for RXX, RYY, RZZ
    """

    theta: float
    _num_target_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=2)


class RXXGate(DoubleRotationGate):
    """
    Rotation by theta about the XX axis
    """


class RYYGate(DoubleRotationGate):
    """
    Rotation by theta about the YY axis
    """


class RZZGate(DoubleRotationGate):
    """
    Rotation by theta about the ZZ axis
    """



from typing import TYPE_CHECKING, List, Optional

import pydantic

# This file is based on autogenerated code from: https://static.ionq.co/schemas/circuit-v0.json using
# https://pydantic-docs.helpmanual.io/datamodel_code_generator/
# Run: datamodel-codegen --url https://static.ionq.co/schemas/circuit-v0.json

if TYPE_CHECKING:
    pydanticGateName = str
else:
    pydanticGateName = pydantic.constr(
        regex=r"^\w+$",
        min_length=1,
    )


class Gate(pydantic.BaseModel):
    gate: pydanticGateName
    target: Optional[int] = None
    control: Optional[int] = None
    targets: Optional[List[int]] = None
    controls: Optional[List[int]] = None


class IonqQuantumCircuit(pydantic.BaseModel):
    qubits: int
    circuit: List[Gate]



from typing import Tuple

import pydantic

from classiq_interface.helpers.custom_pydantic_types import pydanticNonEmptyString

QubitsType = Tuple[pydantic.NonNegativeInt, ...]


class Register(pydantic.BaseModel):
    """
    A user-defined custom register.
    """

    name: pydanticNonEmptyString = pydantic.Field(
        description="The name of the custom register",
    )

    qubits: QubitsType = pydantic.Field(
        description="A tuple of qubits as integers as indexed within a custom function code",
    )

    @property
    def width(self) -> pydantic.PositiveInt:
        """The number of qubits of the custom register"""
        return len(self.qubits)

    @pydantic.validator("qubits")
    def validate_qubits(cls, qubits: QubitsType):
        if len(qubits) == 0:
            raise ValueError("qubits field must be non-empty.")
        if len(set(qubits)) != len(qubits):
            raise ValueError("All qubits of a register must be distinct.")
        return qubits



from typing import TypeVar

import pydantic

Qubit = TypeVar("Qubit", bound=pydantic.NonNegativeInt)
Cycle = TypeVar("Cycle", bound=pydantic.NonNegativeInt)
Clbit = TypeVar("Clbit", bound=pydantic.NonNegativeInt)



from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import pydantic

import classiq_interface.generator.validations.flow_graph as flow_graph
from classiq_interface.generator.function_call import FunctionCall
from classiq_interface.generator.functions import FunctionLibraryData, FunctionType
from classiq_interface.generator.preferences.randomness import create_random_seed
from classiq_interface.generator.qiskit_quantum_gates import QiskitBuiltinQuantumGates
from classiq_interface.generator.range_types import NonNegativeIntRange
from classiq_interface.generator.result import QuantumFormat
from classiq_interface.generator.user_defined_function_params import CustomFunction

DEFAULT_MINIMAL_DEPTH = 1


if TYPE_CHECKING:
    pydanticConstrainedQuantumFormatList = List[QuantumFormat]
else:
    pydanticConstrainedQuantumFormatList = pydantic.conlist(
        QuantumFormat, min_items=1, max_items=len(QuantumFormat)
    )


# TODO define a type that can be used in variable declaration that is consistent with usage
def normalize_dict_key_to_str(d: Dict[Any, Any]) -> Dict[str, Any]:
    return {k.name: v for k, v in d.items()}


class QuantumCircuitConstraints(pydantic.BaseModel):
    """
    Input constraints for the generated quantum circuit.
    """

    # TODO: Consider moving timeout outside of constraints, and supply it (optionally) separate of the constraints.
    # TODO: Remove hard coded timeout when issue,https://github.com/MiniZinc/minizinc-python/pull/8 is resolved
    timeout_seconds: pydantic.PositiveInt = pydantic.Field(
        default=300, description="Generation timeout in seconds"
    )
    qubit_count: Optional[pydantic.PositiveInt] = pydantic.Field(
        default=None,
        description="Number of qubits in generated quantum circuit",
    )
    max_depth: Optional[pydantic.PositiveInt] = None
    random_seed: int = pydantic.Field(
        default_factory=create_random_seed,
        description="The random seed used for the generation",
    )
    gate_count_constraints: Dict[
        QiskitBuiltinQuantumGates, NonNegativeIntRange
    ] = pydantic.Field(default_factory=lambda: defaultdict(NonNegativeIntRange))
    # Must be validated before logic_flow
    function_library: Optional[FunctionLibraryData] = pydantic.Field(
        default=None,
        description="The user-defined custom function library.",
    )
    logic_flow: List[FunctionCall] = pydantic.Field(
        default_factory=list,
        description="List of function calls to be applied in the circuit",
    )
    output_format: Union[
        QuantumFormat, pydanticConstrainedQuantumFormatList
    ] = pydantic.Field(
        default=[QuantumFormat.QASM],
        description="The quantum circuit output format(s). "
        "When multiple formats are requested, only the first one will be presented in "
        "the VSCode extension.",
    )

    _gate_count_constraints = pydantic.validator(
        "gate_count_constraints", allow_reuse=True
    )(normalize_dict_key_to_str)

    draw_as_functions: bool = pydantic.Field(
        default=True,
        description="If true, the generation output will be "
        "visualized as functions and not as an unrolled circuit",
    )
    transpile_circuit: bool = pydantic.Field(
        default=True,
        description="If true, the returned result will contain a "
        "transpiled circuit and its depth",
    )

    class Config:
        extra = "forbid"

    @pydantic.validator("logic_flow")
    def validate_logic_flow(
        cls, logic_flow: List[FunctionCall], values: Dict[str, Any]
    ) -> List[FunctionCall]:
        if not logic_flow:
            return logic_flow

        functions_to_validate = logic_flow.copy()
        library = values.get("function_library")

        while functions_to_validate:
            function_call = functions_to_validate.pop()
            params = function_call.function_params
            if not isinstance(params, CustomFunction):
                continue

            FunctionLibraryData.validate_function_in_library(
                library=library, function_params=params
            )
            assert isinstance(library, FunctionLibraryData)
            function_data = library.function_dict[params.name]
            params.generate_io_names(
                input_set=function_data.input_set,
                output_set=function_data.output_set,
            )
            function_call.validate_custom_function_io()
            if function_data.function_type == FunctionType.CompositeFunction:
                functions_to_validate.extend(function_data.logic_flow)

        flow_graph.validate_flow_graph(logic_flow)

        return logic_flow

    @pydantic.validator("output_format")
    def validate_output_format(cls, output_format):
        if isinstance(output_format, QuantumFormat):
            return [output_format]
        else:
            if len(output_format) == len(set(output_format)):
                return output_format
            else:
                raise ValueError(
                    f"{output_format=}\n"
                    "has at least one format that appears twice or more"
                )



from typing import List

import numpy as np
import pyomo.core as pyo


def ilp(coeffs: List[int], bound: int) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    model.x = pyo.Var(
        range(len(coeffs)), domain=pyo.NonNegativeIntegers, bounds=(0, bound)
    )

    @model.Constraint(range(len(coeffs) - 1))
    def monotone_rule(model, idx):
        return model.x[idx] <= model.x[idx + 1]

    model.cost = pyo.Objective(
        expr=coeffs @ np.array(model.x.values()), sense=pyo.maximize
    )

    return model



from typing import Set, Type

from classiq_interface.generator.amplitude_estimation import AmplitudeEstimation
from classiq_interface.generator.arith.arithmetic import Arithmetic, ArithmeticOracle
from classiq_interface.generator.arith.binary_ops import (
    Adder,
    BitwiseAnd,
    BitwiseOr,
    BitwiseXor,
    CyclicShift,
    Equal,
    GreaterEqual,
    GreaterThan,
    LessEqual,
    LessThan,
    LShift,
    Multiplier,
    NotEqual,
    RShift,
    Subtractor,
)
from classiq_interface.generator.arith.unary_ops import BitwiseInvert, Negation
from classiq_interface.generator.credit_risk_example.cdf_comparator import CDFComparator
from classiq_interface.generator.credit_risk_example.linear_gci import LinearGCI
from classiq_interface.generator.credit_risk_example.weighted_adder import WeightedAdder
from classiq_interface.generator.entangler_params import (
    GridEntangler,
    HypercubeEntangler,
    TwoDimensionalEntangler,
)
from classiq_interface.generator.finance import Finance, FinanceModels, FinancePayoff
from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.generator.grover_operator import GroverOperator
from classiq_interface.generator.hadamard_amp_load import HadamardAmpLoad
from classiq_interface.generator.hardware_efficient_ansatz import (
    HardwareEfficientAnsatz,
)
from classiq_interface.generator.hartree_fock import HartreeFock
from classiq_interface.generator.linear_pauli_rotations import LinearPauliRotations
from classiq_interface.generator.mcx import Mcx
from classiq_interface.generator.qft import QFT
from classiq_interface.generator.sparse_amp_load import SparseAmpLoad
from classiq_interface.generator.standard_gates.standard_gates_param_list import (
    get_qiskit_standard_function_param_list,
)
from classiq_interface.generator.state_preparation import StatePreparation
from classiq_interface.generator.state_propagator import StatePropagator
from classiq_interface.generator.unitary_gate import UnitaryGate
from classiq_interface.generator.user_defined_function_params import CustomFunction

_function_param_list = {
    StatePreparation,
    StatePropagator,
    QFT,
    BitwiseAnd,
    BitwiseOr,
    BitwiseXor,
    BitwiseInvert,
    Adder,
    Arithmetic,
    ArithmeticOracle,
    Equal,
    NotEqual,
    GreaterThan,
    GreaterEqual,
    LessThan,
    LessEqual,
    Negation,
    Subtractor,
    RShift,
    LShift,
    CyclicShift,
    TwoDimensionalEntangler,
    Finance,
    FinanceModels,
    FinancePayoff,
    HypercubeEntangler,
    AmplitudeEstimation,
    SparseAmpLoad,
    GridEntangler,
    HadamardAmpLoad,
    GroverOperator,
    Mcx,
    CustomFunction,
    HardwareEfficientAnsatz,
    UnitaryGate,
    WeightedAdder,
    LinearPauliRotations,
    Multiplier,
    LinearGCI,
    CDFComparator,
    HartreeFock,
}


def get_function_param_list() -> Set[Type[FunctionParams]]:
    qiskit_standard_function_param_list = get_qiskit_standard_function_param_list()
    _function_param_list.update(qiskit_standard_function_param_list)
    return _function_param_list



from typing import Dict, Optional

import pydantic

from classiq_interface.backend.backend_preferences import (
    BackendPreferencesTypes,
    backend_preferences_field,
)
from classiq_interface.combinatorial_optimization.encoding_types import EncodingType
from classiq_interface.combinatorial_optimization.preferences import QAOAPreferences
from classiq_interface.executor.execution_preferences import OptimizerPreferences


class OptimizationProblem(pydantic.BaseModel):
    qaoa_preferences: QAOAPreferences = pydantic.Field(
        default_factory=QAOAPreferences,
        description="preferences for the QAOA algorithm",
    )
    optimizer_preferences: OptimizerPreferences = pydantic.Field(
        default_factory=OptimizerPreferences,
        description="preferences for the VQE execution",
    )
    serialized_model: Optional[Dict] = None
    backend_preferences: BackendPreferencesTypes = backend_preferences_field()
    encoding_type: Optional[EncodingType] = pydantic.Field(
        default_factory=EncodingType.BINARY,
        description="encoding scheme for integer variables",
    )



import pydantic

from classiq_interface.chemistry.ground_state_problem import GroundStateProblem
from classiq_interface.executor.execution_preferences import ExecutionPreferences
from classiq_interface.generator.constraints import QuantumCircuitConstraints


class GroundStateProblemExecution(pydantic.BaseModel):
    molecule: GroundStateProblem
    quantum_circuit_constraints: QuantumCircuitConstraints
    preferences: ExecutionPreferences



import ast
import enum
import keyword
import random
import re
from enum import Enum
from typing import Dict, Optional, Union

import pydantic

from classiq_interface.generator.arith.fix_point_number import (
    MAX_FRACTION_PLACES,
    FixPointNumber,
)
from classiq_interface.generator.arith.register_user_input import RegisterUserInput
from classiq_interface.generator.function_params import FunctionParams

DEFAULT_OUT_NAME = "out"
DEFAULT_ARG_NAME = "in_arg"

SUPPORTED_FUNC_NAMES = ("CLShift", "CRShift")

white_list = {"or", "and"}.union(SUPPORTED_FUNC_NAMES)
black_list = set(keyword.kwlist) - white_list


class MappingMethods(str, enum.Enum):
    topological = "topological"
    pebble = "pebble"


class Arithmetic(FunctionParams):
    max_fraction_places: pydantic.NonNegativeInt = MAX_FRACTION_PLACES
    expression: str
    definitions: Dict[
        str,
        Union[
            pydantic.StrictInt, pydantic.StrictFloat, FixPointNumber, RegisterUserInput
        ],
    ]
    method: MappingMethods = MappingMethods.pebble
    qubit_count: Optional[pydantic.NonNegativeInt] = None
    output_name: Optional[str]
    random_seed: Optional[int] = None

    @pydantic.validator("expression")
    def check_expression_is_legal(cls, expression):
        ast.parse(expression, "", "eval")
        return expression

    @pydantic.root_validator()
    def check_all_variable_are_defined(cls, values):
        expression, definitions = values.get("expression"), values.get("definitions")

        literals = set(re.findall("[A-Za-z][A-Za-z0-9]*", expression))

        not_allowed = literals.intersection(black_list)
        undefined_literals = literals.difference(definitions, white_list)
        if not_allowed:
            raise ValueError(f"The following names: {not_allowed} are not allowed")

        if undefined_literals:
            raise ValueError(f"{undefined_literals} need to be defined in definitions")
        return values

    @pydantic.root_validator()
    def substitute_expression(cls, values):
        # TODO there isn't a secure way to simplify the expression which does not involve using eval.
        #  Can be done with sdk on client side
        expression, definitions = values.get("expression"), values.get("definitions")
        new_definition = dict()
        for var_name, value in definitions.items():
            if isinstance(value, RegisterUserInput):
                new_definition[var_name] = value
                continue
            elif isinstance(value, int):
                pass
            elif isinstance(value, float):
                value = FixPointNumber(float_value=value).actual_float_value
            elif isinstance(value, FixPointNumber):
                value = value.actual_float_value
            else:
                raise ValueError(f"{type(value)} type is illegal")

            expression = re.sub(r"\b" + var_name + r"\b", str(value), expression)
        values["expression"] = expression
        values["definitions"] = new_definition
        return values

    @pydantic.validator("definitions")
    def set_register_names(cls, definitions):
        for k, v in definitions.items():
            if isinstance(v, RegisterUserInput):
                v.name = k
        return definitions

    @pydantic.validator("random_seed", always=True)
    def validate_random_seed(cls, random_seed):
        return random.randint(0, 1000) if not random_seed else random_seed

    def create_io_enums(self):
        literals = set(re.findall("[A-Za-z][A-Za-z0-9]*", self.expression))
        output_name = self.output_name if self.output_name else DEFAULT_OUT_NAME
        self._input_enum = Enum(
            "ArithmeticInputs",
            {literal: literal for literal in literals if literal not in white_list},
        )
        self._output_enum = Enum("ArithmeticOutputs", {output_name: output_name})

    class Config:
        extra = "forbid"


class ArithmeticOracle(Arithmetic):
    @pydantic.validator("expression")
    def validate_compare_expression(cls, value):
        ast_obj = ast.parse(value, "", "eval")
        if not isinstance(ast_obj.body, (ast.Compare, ast.BoolOp)):
            raise ValueError("Must a comparison expression")

        return value



from typing import TYPE_CHECKING, Any, List

import pydantic

# General int types

if TYPE_CHECKING:
    pydanticLargerThanOneInteger = int
else:

    class pydanticLargerThanOneInteger(pydantic.ConstrainedInt):
        gt = 1


# Probability float types
if TYPE_CHECKING:
    pydanticProbabilityFloat = float
    pydanticNonOneProbabilityFloat = float
    pydanticNonZeroProbabilityFloat = float
else:

    class pydanticProbabilityFloat(pydantic.ConstrainedFloat):
        ge = 0.0
        le = 1.0

    class pydanticNonOneProbabilityFloat(pydantic.ConstrainedFloat):
        ge = 0.0
        lt = 1.0

    class pydanticNonZeroProbabilityFloat(pydantic.ConstrainedFloat):
        gt = 0.0
        le = 1.0


# CVAR parameter types
if TYPE_CHECKING:
    pydanticAlphaParamCVAR = float
else:

    class pydanticAlphaParamCVAR(pydantic.ConstrainedFloat):
        gt = 0.0
        le = 1.0


# General string types
if TYPE_CHECKING:
    pydanticNonEmptyString = str
else:
    pydanticNonEmptyString = pydantic.constr(min_length=1)

# Name string types
if TYPE_CHECKING:
    pydanticFunctionNameStr = str
else:
    pydanticFunctionNameStr = pydantic.constr(
        strict=True, regex="^([a-z][a-z0-9]*)(_[a-z0-9]+)*$"
    )

if TYPE_CHECKING:
    pydanticPauliMonomialStr = str
else:
    pydanticPauliMonomialStr = pydantic.constr(
        strict=True, strip_whitespace=True, min_length=1, regex="^[IXYZ]+$"
    )

if TYPE_CHECKING:
    pydanticPauliList = List[list]
else:
    pydanticPauliList = pydantic.conlist(
        item_type=pydantic.conlist(item_type=Any, min_items=2, max_items=2),
        min_items=1,
    )



from typing import Optional

import pydantic

from classiq_interface.helpers.custom_pydantic_types import pydanticProbabilityFloat


class NoiseProperties(pydantic.BaseModel):

    measurement_bit_flip_probability: Optional[
        pydanticProbabilityFloat
    ] = pydantic.Field(
        default=None,
        description="Probability of measuring the wrong value for each qubit.",
    )



from datetime import timedelta
from typing import TYPE_CHECKING

import pydantic

AZURE_QUANTUM_RESOURCE_ID_REGEX = r"^/subscriptions/([a-fA-F0-9-]*)/resourceGroups/([^\s/]*)/providers/Microsoft\.Quantum/Workspaces/([^\s/]*)$"  # noqa: F722

_IONQ_API_KEY_LENGTH: int = 32
INVALID_API_KEY: str = _IONQ_API_KEY_LENGTH * "a"
MAX_EXECUTION_TIMEOUT_SECONDS = timedelta(hours=4).total_seconds()

if TYPE_CHECKING:
    pydanticAWSAccessKeyID = str
    pydanticAWSSecretAccessKey = str
    pydanticExecutionTimeout = int
    pydanticRegionName = str
    pydanticS3BucketKey = str
    pydanticS3BucketName = str
    pydanticAzureResourceIDType = str
    pydanticIonQApiKeyType = str
    pydanticArgumentNameType = str
else:
    # TODO Simplify regular expressions in this file

    pydanticAWSAccessKeyID = pydantic.constr(
        strip_whitespace=True,
        min_length=20,
        max_length=20,
        regex=r"(?<![A-Z0-9])[A-Z0-9]{20}(?![A-Z0-9])",
    )

    pydanticAWSSecretAccessKey = pydantic.constr(
        strip_whitespace=True,
        min_length=40,
        max_length=40,
        regex=r"(?<![A-Za-z0-9/+=])[A-Za-z0-9/+=]{40}(?![A-Za-z0-9/+=])",
    )

    pydanticRegionName = pydantic.constr(strip_whitespace=True, min_length=1)

    pydanticS3BucketName = pydantic.constr(strip_whitespace=True, min_length=3)

    pydanticS3BucketKey = pydantic.constr(strip_whitespace=True, min_length=1)

    pydanticAzureResourceIDType = pydantic.constr(regex=AZURE_QUANTUM_RESOURCE_ID_REGEX)

    pydanticIonQApiKeyType = pydantic.constr(
        regex=f"[A-Za-z0-9]{{{_IONQ_API_KEY_LENGTH}}}"
    )
    pydanticExecutionTimeout = pydantic.conint(gt=0, le=MAX_EXECUTION_TIMEOUT_SECONDS)

    pydanticArgumentNameType = pydantic.constr(regex="[_a-zA-Z][_a-zA-Z0-9]*")



ROUTE_PREFIX = "/api/v1"
ANALYZER_DASHBOARD = "/analyzer"
ANALYZER_PREFIX = ROUTE_PREFIX + "/analyzer"
CHEMISTRY_PREFIX = ROUTE_PREFIX + "/chemistry"
EXECUTE_PREFIX = ROUTE_PREFIX + "/execute"
COMBINATORIAL_OPTIMIZATION_PREFIX = ROUTE_PREFIX + "/combinatorial_optimization"

TASKS_SUFFIX = "/tasks"

TASKS_GENERATE_SUFFIX = "/tasks/generate"
TASKS_SOLVE_SUFFIX = "/tasks/solve"
TASKS_WS_SOLVE_SUFFIX = "/ws/tasks/solve"
TASKS_OPERATOR_SUFFIX = "/tasks/operator"
TASKS_OBJECTIVE_SUFFIX = "/tasks/objective"
TASKS_INITIAL_POINT_SUFFIX = "/tasks/initial_point"

EXECUTE_TASKS_FULL_PATH = EXECUTE_PREFIX + TASKS_SUFFIX

ANALYZER_FULL_PATH = ANALYZER_PREFIX + TASKS_SUFFIX
COMBINATORIAL_OPTIMIZATION_GENERATE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_GENERATE_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_SOLVE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_SOLVE_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_WS_SOLVE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_WS_SOLVE_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_OPERATOR_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_OPERATOR_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_OBJECTIVE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_OBJECTIVE_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_INITIAL_POINT_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_INITIAL_POINT_SUFFIX
)
GENERATE_CHEMISTRY_HAMILTONIAN_FULL_PATH = CHEMISTRY_PREFIX + TASKS_SUFFIX



import enum
from itertools import chain
from typing import Any, List, Optional, Set, Tuple, Union

import pydantic

from classiq_interface.generator import function_call
from classiq_interface.generator.functions.function_implementation import (
    FunctionImplementation,
    Register,
    to_tuple,
)
from classiq_interface.helpers.custom_pydantic_types import pydanticFunctionNameStr

ImplementationsType = Union[Tuple[FunctionImplementation, ...], FunctionImplementation]


class FunctionType(str, enum.Enum):
    ElementaryFunction = "ElementaryFunction"
    CompositeFunction = "CompositeFunction"


def _first_impl(impl_type: ImplementationsType) -> FunctionImplementation:
    if isinstance(impl_type, FunctionImplementation):
        return impl_type
    else:
        return impl_type[0]


class FunctionData(pydantic.BaseModel):
    """
    Facilitates the creation of a user-defined custom function
    """

    name: pydanticFunctionNameStr = pydantic.Field(
        description="The name of a custom function"
    )

    implementations: Optional[ImplementationsType] = pydantic.Field(
        description="The implementations of the custom function",
    )
    logic_flow: List[function_call.FunctionCall] = pydantic.Field(
        default=list(), description="List of function calls to perform."
    )

    _function_type: FunctionType = pydantic.PrivateAttr(default=None)
    _input_set: Set[str] = pydantic.PrivateAttr(default_factory=set)
    _output_set: Set[str] = pydantic.PrivateAttr(default_factory=set)

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.logic_flow:
            self._function_type = FunctionType.CompositeFunction
            self._input_set, self._output_set = self._deduce_composite_io_names(
                self.logic_flow
            )
            return

        inputs_regs: Tuple[Register, ...] = (
            to_tuple(_first_impl(self.implementations).input_registers)
            if self.implementations is not None
            else tuple()
        )
        output_regs: Tuple[Register, ...] = (
            to_tuple(_first_impl(self.implementations).output_registers)
            if self.implementations is not None
            else tuple()
        )

        self._function_type = FunctionType.ElementaryFunction
        self._input_set = set(register.name for register in inputs_regs)
        self._output_set = set(register.name for register in output_regs)

    @staticmethod
    def _deduce_composite_io_names(
        logic_flow: List[function_call.FunctionCall],
    ) -> Tuple[Set[str], Set[str]]:
        input_wires = list(
            chain(*[call.inputs.values() for call in logic_flow if call.inputs])
        )
        output_wires = list(
            chain(*[call.outputs.values() for call in logic_flow if call.outputs])
        )

        input_names = {name for name in input_wires if name not in output_wires}
        output_names = {name for name in output_wires if name not in input_wires}

        return input_names, output_names

    @property
    def input_set(self) -> Set[str]:
        return self._input_set

    @property
    def output_set(self) -> Set[str]:
        return self._output_set

    @property
    def function_type(self) -> FunctionType:
        return self._function_type

    @pydantic.validator("name")
    def validate_name(cls, name: str):
        validate_name_end_not_newline(name=name)
        return name

    @pydantic.validator("implementations")
    def validate_implementations(cls, implementations: ImplementationsType):
        if implementations is None:
            return implementations

        if not implementations:
            raise ValueError(
                "The implementations of a custom function must be non-empty."
            )

        if isinstance(implementations, FunctionImplementation):
            implementations = (implementations,)

        distinct_io_signatures = set(
            tuple(
                tuple(
                    sorted(
                        (register.name, register.width)
                        for register in to_tuple(registers)
                    )
                )
                for registers in (
                    implementation.input_registers,
                    implementation.output_registers,
                )
            )
            for implementation in implementations
        )
        if len(distinct_io_signatures) != 1:
            raise ValueError(
                "All implementations of a custom function must have matching IO "
                "register names and widths."
            )
        return implementations

    @pydantic.root_validator
    def validate_logic_flow(cls, values):
        implementations = values.get("implementations")
        logic_flow = values.get("logic_flow")
        elementary = bool(implementations is not None)
        composite = bool(len(logic_flow) > 0)
        if elementary == composite:
            raise ValueError(
                "Function must contain either implementations or calls, but not both"
            )

        return values


def validate_name_end_not_newline(name: str):
    _NEW_LINE = "\n"
    if name.endswith(_NEW_LINE):
        raise ValueError("Function name cannot end in a newline character")



import enum
from typing import Any, Dict, Optional, Union

import pydantic

from classiq_interface.executor.vqe_result import VQESolverResult


class ExecutionStatus(str, enum.Enum):
    SUCCESS = "success"
    ERROR = "error"


class VaRResult(pydantic.BaseModel):
    var: Optional[float] = None
    alpha: Optional[float] = None


class FinanceSimulationResults(pydantic.BaseModel):
    var_results: Optional[VaRResult] = None
    result: Optional[float] = None

    @pydantic.root_validator()
    def validate_atleast_one_field(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        is_var_results_defined = values.get("var_results") is not None
        is_result_defined = values.get("result") is not None

        if not is_var_results_defined and not is_result_defined:
            raise ValueError(
                "At least one of var_results and result should be defined."
            )

        return values


class GroverSimulationResults(pydantic.BaseModel):
    result: Dict[str, Union[float, int]]


class ExecutionDetails(pydantic.BaseModel):
    vendor_format_result: Dict[str, Any] = pydantic.Field(
        ..., description="Result in proprietary vendor format"
    )
    counts: Dict[str, pydantic.NonNegativeInt] = pydantic.Field(
        default_factory=dict, description="Number of counts per state"
    )
    histogram: Optional[Dict[str, pydantic.NonNegativeFloat]] = pydantic.Field(
        None,
        description="Histogram of probability per state (an alternative to counts)",
    )


ExecutionData = Union[
    ExecutionDetails,
    FinanceSimulationResults,
    GroverSimulationResults,
    VQESolverResult,
    str,
]


class ExecutionResult(pydantic.BaseModel):
    status: ExecutionStatus
    details: ExecutionData



import pydantic

from classiq_interface.chemistry.operator import PauliOperator
from classiq_interface.executor.quantum_program import QuantumProgram


class HamiltonianMinimizationProblem(pydantic.BaseModel):
    ansatz: QuantumProgram
    hamiltonian: PauliOperator



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Optional, Union

import pydantic

from classiq_interface.backend.ionq import ionq_quantum_program
from classiq_interface.backend.pydantic_backend import pydanticArgumentNameType
from classiq_interface.executor.quantum_instruction_set import QuantumInstructionSet

Arguments = Optional[Dict[pydanticArgumentNameType, Any]]


class QuantumProgram(pydantic.BaseModel):
    syntax: QuantumInstructionSet = pydantic.Field(
        ..., description="The syntax of the program."
    )
    code: Union[str, ionq_quantum_program.IonqQuantumCircuit] = pydantic.Field(
        ..., description="The textual representation of the program"
    )
    arguments: Arguments = pydantic.Field(
        None,
        description="The parameters dictionary for a parametrized quantum program. "
        "Relevant for Q# programs only.",
    )

    @pydantic.validator("code")
    def load_quantum_program(cls, code: str, values: Dict[str, Any]):
        if not isinstance(code, str):
            return code

        syntax = values.get("syntax")
        if syntax == QuantumInstructionSet.IONQ:
            return ionq_quantum_program.IonqQuantumCircuit.parse_raw(code)

        return code

    @pydantic.validator("arguments")
    def validate_arguments(cls, arguments: Arguments, values: Dict[str, Any]):
        if arguments and values.get("syntax") != QuantumInstructionSet.QSHARP:
            raise ValueError("Only Q# programs support arguments")
        return arguments

    @staticmethod
    def from_file(
        file_path: Union[str, Path],
        syntax: Optional[Union[str, QuantumInstructionSet]] = None,
        arguments: Optional[Arguments] = None,
    ) -> QuantumProgram:
        path = Path(file_path)
        code = path.read_text()
        if syntax is None:
            syntax = path.suffix.lstrip(".")
        return QuantumProgram(syntax=syntax, code=code, arguments=arguments)



from classiq_interface.generator.functions.function_data import *
from classiq_interface.generator.functions.function_implementation import *
from classiq_interface.generator.functions.function_library_data import *
from classiq_interface.generator.functions.register import *



from typing import Any, Optional, Union

import pydantic

from classiq_interface.generator.standard_gates.standard_gates import StandardGate

CONTROLLED_GATE_CONTROL_INPUT = "CTRL_IN"
CONTROLLED_GATE_TARGET_INPUT = "TARGET_IN"

CONTROLLED_GATE_CONTROL_OUTPUT = "CTRL_OUT"
CONTROLLED_GATE_TARGET_OUTPUT = "TARGET_OUT"

CtrlState = Optional[Union[pydantic.StrictStr, pydantic.NonNegativeInt]]


class ControlledGate(StandardGate):
    """
    Base model for controlled Gates
    """

    _num_ctrl_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=1)

    _input_names = pydantic.PrivateAttr(
        default=[CONTROLLED_GATE_CONTROL_INPUT, CONTROLLED_GATE_TARGET_INPUT]
    )
    _output_names = pydantic.PrivateAttr(
        default=[CONTROLLED_GATE_CONTROL_OUTPUT, CONTROLLED_GATE_TARGET_OUTPUT]
    )


class ControlledGateWithState(ControlledGate):
    """
    Base model for controlled Gates with control over the controlled_state
    """

    ctrl_state: CtrlState = pydantic.Field(
        description="The control state in decimal or as a bit string (e.g. ‘1011’). If not specified, the control "
        "state is 2**num_ctrl_qubits - 1 "
    )

    def __init__(self, **data: Any):
        super().__init__(**data)

        self.validate_ctrl_state()

    def validate_ctrl_state(self) -> None:
        num_ctrl_qubits = self._num_ctrl_qubits
        if self.ctrl_state is None:
            self.ctrl_state = pydantic.StrictStr("1" * num_ctrl_qubits)
            return
        ctrl_state_int: pydantic.NonNegativeInt = (
            int(self.ctrl_state, 2)
            if isinstance(self.ctrl_state, str)
            else self.ctrl_state
        )
        if ctrl_state_int < 0 or ctrl_state_int >= 2 ** num_ctrl_qubits:
            raise ValueError(
                "Control state value should be zero or positive and smaller than 2**num_ctrl_qubits"
            )


class CXGate(ControlledGateWithState):
    """
    The Controlled-X Gate
    """


class CCXGate(ControlledGateWithState):
    """
    The Double Controlled-X Gate
    """

    _num_ctrl_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=2)


class C3XGate(ControlledGateWithState):
    """
    The X Gate controlled on 3 qubits
    """

    _num_ctrl_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=3)


class C4XGate(ControlledGateWithState):
    """
    The X Gate controlled on 4 qubits
    """

    _num_ctrl_qubits: pydantic.PositiveInt = pydantic.PrivateAttr(default=4)


class CYGate(ControlledGateWithState):
    """
    The Controlled-Y Gate
    """


class CZGate(ControlledGateWithState):
    """
    The Controlled-Z Gate
    """


class CHGate(ControlledGateWithState):
    """
    The Controlled-H Gate
    """


class CSXGate(ControlledGateWithState):
    """
    The Controlled-SX Gate
    """


class CRXGate(ControlledGateWithState):
    """
    The Controlled-RX Gate
    """

    theta: float


class CRYGate(ControlledGateWithState):
    """
    The Controlled-RY Gate
    """

    theta: float


class CRZGate(ControlledGateWithState):
    """
    The Controlled-RZ Gate
    """

    theta: float


class CPhaseGate(ControlledGateWithState):
    """
    The Controlled-Phase Gate
    """

    theta: float


class MCPhaseGate(ControlledGate):
    """
    The Controlled-Phase Gate
    """

    num_ctrl_qubits: pydantic.PositiveInt

    lam: float

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._num_ctrl_qubits = self.num_ctrl_qubits



from __future__ import annotations

import random
import string
from typing import Dict, KeysView, List, Optional

import pydantic

from classiq_interface.generator import function_param_list, function_params as f_params
from classiq_interface.generator.user_defined_function_params import CustomFunction
from classiq_interface.helpers.custom_pydantic_types import pydanticNonEmptyString

_SUFFIX_LEN = 6
BAD_FUNCTION_ERROR_MSG = "Unknown function"
BAD_INPUT_ERROR_MSG = "Bad input name given"
BAD_OUTPUT_ERROR_MSG = "Bad output name given"


class FunctionCall(pydantic.BaseModel):
    function: str = pydantic.Field(
        default="", description="The function that is called"
    )
    function_params: f_params.FunctionParams = pydantic.Field(
        description="The parameters necessary for defining the function"
    )
    is_inverse: bool = pydantic.Field(
        default=False, description="call to function inverse"
    )
    inputs: Dict[pydanticNonEmptyString, pydanticNonEmptyString] = pydantic.Field(
        default_factory=dict,
        description="A mapping from the input name to the wire it connects to",
    )
    outputs: Dict[pydanticNonEmptyString, pydanticNonEmptyString] = pydantic.Field(
        default_factory=dict,
        description="A mapping from the output name to the wire it connects to",
    )

    name: Optional[pydanticNonEmptyString] = pydantic.Field(
        default=None,
        description="The name of the function call. Determined automatically.",
    )

    @pydantic.validator("name", always=True)
    def create_name(cls, name, values):
        function = values.get("function")
        params = values.get("function_params")
        if not function or params is None:
            return name

        suffix = "".join(
            random.choice(string.ascii_letters + string.digits)
            for _ in range(_SUFFIX_LEN)
        )
        if isinstance(params, CustomFunction):
            return f"{params.name}_{suffix}"
        return f"{function}_{suffix}"

    @pydantic.validator("function_params", pre=True)
    def parse_function_params(cls, function_params, values):
        if isinstance(function_params, f_params.FunctionParams):
            values["function"] = type(function_params).__name__
            return function_params

        function = values.get("function")
        if not function:
            raise ValueError(
                "The function field must be provided to deduce function type"
            )

        func_class = [
            seg
            for seg in function_param_list.get_function_param_list()
            if seg.__name__ == function
        ]

        if not func_class:
            raise ValueError(f"{BAD_FUNCTION_ERROR_MSG}: {function}")

        return func_class[0].parse_obj(function_params)

    @staticmethod
    def _validate_input_names(
        params: f_params.FunctionParams, inputs: Dict[str, str], is_inverse: bool
    ) -> None:
        invalid_names = FunctionCall._get_invalid_io_names(
            inputs.keys(), params, f_params.input_io(is_inverse)
        )
        if invalid_names:
            raise ValueError(f"{BAD_INPUT_ERROR_MSG}: {invalid_names}")

    @pydantic.validator("inputs")
    def validate_inputs(cls, inputs, values):
        params = values.get("function_params")
        is_inverse = values.get("is_inverse")
        if params is None:
            return inputs
        if isinstance(params, CustomFunction):
            return inputs
        cls._validate_input_names(params=params, inputs=inputs, is_inverse=is_inverse)
        return inputs

    @staticmethod
    def _validate_output_names(
        params: f_params.FunctionParams, outputs: Dict[str, str], is_inverse: bool
    ) -> None:
        invalid_names = FunctionCall._get_invalid_io_names(
            outputs.keys(), params, f_params.output_io(is_inverse)
        )
        if invalid_names:
            raise ValueError(f"{BAD_OUTPUT_ERROR_MSG}: {invalid_names}")

    @pydantic.validator("outputs")
    def validate_outputs(cls, outputs, values):
        params = values.get("function_params")
        is_inverse = values.get("is_inverse")
        if params is None:
            return outputs
        if isinstance(params, CustomFunction):
            return outputs
        cls._validate_output_names(
            params=params, outputs=outputs, is_inverse=is_inverse
        )
        return outputs

    @staticmethod
    def _get_invalid_io_names(
        names: KeysView[str], params: f_params.FunctionParams, io: f_params.IO
    ) -> List[str]:
        return [name for name in names if not params.is_valid_io_name(name, io)]

    def validate_custom_function_io(self) -> None:
        if not isinstance(self.function_params, CustomFunction):
            raise AssertionError("CustomFunction object expected.")
        FunctionCall._validate_input_names(
            params=self.function_params, inputs=self.inputs, is_inverse=self.is_inverse
        )
        FunctionCall._validate_output_names(
            params=self.function_params,
            outputs=self.outputs,
            is_inverse=self.is_inverse,
        )

    def inverse(self) -> FunctionCall:
        return FunctionCall(
            function=self.function,
            function_params=self.function_params,
            inputs=self.outputs,
            outputs=self.inputs,
            name=self.name,
            is_inverse=not self.is_inverse,
        )



from typing import List, Optional, Union

import numpy as np
import pydantic

from classiq_interface.generator import complex_type, function_params

DataNumber = Union[complex_type.Complex, float, int]
DataArray = List[List[DataNumber]]

UNITARY_GATE_INPUTS = ["IN"]
UNITARY_GATE_OUTPUTS = ["OUT"]
UNITARY_GATE_CONTROLLED_INPUTS = ["CTRL_IN", "TARGET_IN"]
UNITARY_GATE_CONTROLLED_OUTPUTS = ["CTRL_OUT", "TARGET_OUT"]


class UnitaryGate(function_params.FunctionParams):
    """
    Creates a circuit implementing a specified 2**n * 2**n unitary transformation.
    Use the num_ctrl_qubits to create a controlled unitary transformation, and the ctrl_state field to customize the
    control state.
    """

    # TODO - add support to numpy array-like (requires custom pydantic type definition)
    data: DataArray = pydantic.Field(
        description="A 2**n * 2**n (n positive integer) unitary matrix."
    )
    num_ctrl_qubits: Optional[pydantic.PositiveInt] = pydantic.Field(
        description="If specified and greater than 0, returns a num_ctrl_qubits controlled unitary gate"
    )
    ctrl_state: Optional[Union[str, pydantic.NonNegativeInt]] = pydantic.Field(
        description="The control state in decimal or as a bit string (e.g. ‘1011’). If not specified, the control state is 2**num_ctrl_qubits - 1"
    )

    # TODO - decide if to include assertion on the unitarity of the matrix. It is already done in Qiskit and could be computationally expensive
    @pydantic.validator("data")
    def validate_data(cls, data):
        data_np = np.array(data)
        if data_np.ndim != 2:
            raise ValueError("Data must me two dimensional")
        if data_np.shape[0] != data_np.shape[1]:
            raise ValueError("Matrix must be square")
        if not np.mod(np.log2(data_np.shape[0]), 1) == 0:
            raise ValueError("Matrix dimensions must be an integer exponent of 2")
        return data

    @pydantic.validator("ctrl_state")
    def validate_ctrl_state(cls, ctrl_state, values):
        num_ctrl_qubits = values.get("num_ctrl_qubits")
        if num_ctrl_qubits is None:
            return ctrl_state
        if ctrl_state is None:
            return "1" * num_ctrl_qubits
        ctrl_state_int = (
            int(ctrl_state, 2) if isinstance(ctrl_state, str) else ctrl_state
        )
        if 0 <= ctrl_state_int < 2 ** num_ctrl_qubits:
            return ctrl_state
        else:
            raise ValueError(
                "Control state value should be zero or positive and smaller than 2**num_ctrl_qubits"
            )

    def _create_io_names(self):
        if self.num_ctrl_qubits is None:
            self._input_names = UNITARY_GATE_INPUTS
            self._output_names = UNITARY_GATE_OUTPUTS
        else:
            self._input_names = UNITARY_GATE_CONTROLLED_INPUTS
            self._output_names = UNITARY_GATE_CONTROLLED_OUTPUTS



import pyomo.core
import sympy
from pyomo.core.expr.logical_expr import EqualityExpression

from classiq_interface.pyomo_extension import (
    equality_expression,
    inequality_expression,
    set_pprint,
)

pyomo.core.expr.logical_expr.InequalityExpression.getname = (
    inequality_expression.getname
)

pyomo.core.expr.logical_expr.EqualityExpression.getname = equality_expression.getname

pyomo.core.base.set.Set._pprint_members = staticmethod(set_pprint.pprint_members)

pyomo.core.expr.sympy_tools._operatorMap.update({sympy.LessThan: lambda x, y: x <= y})
pyomo.core.expr.sympy_tools._operatorMap.update(
    {sympy.GreaterThan: lambda x, y: x >= y}
)
pyomo.core.expr.sympy_tools._pyomo_operator_map.update({EqualityExpression: sympy.Eq})



from typing import TYPE_CHECKING, List, Union

import pydantic

from classiq_interface.generator import function_params

if TYPE_CHECKING:
    pydanticNonEmptyNonNegativeIntList = List[int]
else:
    pydanticNonEmptyNonNegativeIntList = pydantic.conlist(
        pydantic.conint(ge=0), min_items=1
    )


class AmplitudeEstimation(function_params.FunctionParams):
    """
    Creates a quantum circuit for ampitude estimation
    Provide the state preparation with a qasm string
    """

    state_preparation: str = pydantic.Field(
        description='The state preparation circuit in qasm format. Replace "..." with '
        "'...'. The total number of qubits is the sum of the state preparation qubits "
        "and `num_eval_qubits`"
    )

    objective_qubits: Union[
        pydanticNonEmptyNonNegativeIntList, pydantic.NonNegativeInt
    ] = pydantic.Field(
        default=0,
        description='The list of "good" qubits. The good states are the ones that have '
        "1's in the positions defined by objective qubits. The indices in this list "
        "must be in the range defined by `state_preparation`",
    )

    num_eval_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits to evaluate on the amplitude estimation. "
        "More evaluation qubits provide a better estimate of the good states' amplitude"
    )

    @pydantic.validator("state_preparation")
    def vscode_qasm2circuit(cls, v):
        VSCODE_QUOTE = "'"
        QASM_QUOTES = '"'
        return v.replace(VSCODE_QUOTE, QASM_QUOTES)



from typing import List, Optional, Tuple

import pydantic


class AtomicLocation(pydantic.BaseModel):
    x: float = pydantic.Field(description="The X location of an atom")
    y: float = pydantic.Field(description="The Y location of an atom")
    z: float = pydantic.Field(description="The Z location of an atom")

    def to_list(self) -> List[float]:
        return [self.x, self.y, self.z]


class Atom(pydantic.BaseModel):
    name: str = pydantic.Field(description="The name of the atom")
    location: AtomicLocation = pydantic.Field(description="The location of the atom")


class Molecule(pydantic.BaseModel):
    atoms: List[Atom] = pydantic.Field(description="List of atoms")
    spin: Optional[pydantic.NonNegativeInt] = pydantic.Field(
        default=1, description="spin of the molecule"
    )
    charge: Optional[pydantic.NonNegativeInt] = pydantic.Field(
        default=0, description="charge of the molecule"
    )

    @property
    def geometry(self) -> List[Tuple[str, List[float]]]:
        return [(atom.name, atom.location.to_list()) for atom in self.atoms]



from enum import Enum
from typing import Dict, List, Union

import pydantic

from classiq_interface.generator.arith.arithmetic import (
    DEFAULT_ARG_NAME,
    DEFAULT_OUT_NAME,
)
from classiq_interface.generator.arith.fix_point_number import FixPointNumber
from classiq_interface.generator.arith.register_user_input import RegisterUserInput
from classiq_interface.generator.function_params import FunctionParams


class LogicalOps(FunctionParams):
    args: List[Union[RegisterUserInput, FixPointNumber, int, float]]
    output_name: str = DEFAULT_OUT_NAME

    @pydantic.validator("args")
    def validate_inputs_sizes(cls, args):
        for arg in args:
            if isinstance(arg, RegisterUserInput) and (
                arg.size != 1 or arg.fraction_places != 0
            ):
                raise ValueError(
                    f"All inputs to logical and must be of size 1 | {arg.name}"
                )

        return args

    @pydantic.validator("args")
    def set_inputs_names(cls, args):
        for i, arg in enumerate(args):
            if isinstance(arg, RegisterUserInput):
                arg.name = arg.name if arg.name else DEFAULT_ARG_NAME + str(i)

        return args

    def create_io_enums(self):
        input_names: Dict[str, str] = {
            arg.name: arg.name
            for arg in self.args
            if isinstance(arg, RegisterUserInput)
        }
        output_names: Dict[str, str] = {self.output_name: self.output_name}
        output_names.update(input_names)  # out of place funcs
        self._input_enum = Enum("LogicalAndInputs", input_names)
        self._output_enum = Enum("BinaryOpOutputs", output_names)

    class Config:
        arbitrary_types_allowed = True


class LogicalAnd(LogicalOps):
    pass


class LogicalOr(LogicalOps):
    pass



import sys


def package_version(package: str) -> str:
    try:
        if sys.version_info >= (3, 8):
            from importlib import metadata
        else:
            import importlib_metadata as metadata
    except ImportError:
        raise RuntimeError(
            "Version number unavailable. importlib-metadata is required for Python 3.7"
        )

    return metadata.version(package)


VERSION = package_version(__name__.split(".")[0])



import enum

import numpy as np
import pydantic
from more_itertools import all_equal

from classiq_interface.generator.complex_type import Complex
from classiq_interface.helpers.custom_pydantic_types import (
    pydanticPauliList,
    pydanticPauliMonomialStr,
)


class PauliOperator(pydantic.BaseModel):
    """
    Specification of a Pauli sum operator.
    """

    pauli_list: pydanticPauliList = pydantic.Field(
        description="A list of lists each containing a pauli string comprised of IXYZ characters and a complex coefficient.",
    )

    def show(self) -> str:
        if self.to_hermitian():
            return "\n".join(
                f"{summand[1].real:+.3f} * {summand[0]}" for summand in self.pauli_list
            )
        return "\n".join(
            f"+({summand[1]:+.3f}) * {summand[0]}" for summand in self.pauli_list
        )

    def to_hermitian(self) -> bool:
        if not all(
            np.isclose(complex(summand[1]).real, summand[1])
            for summand in self.pauli_list
        ):
            return False
        self.pauli_list = [
            [summand[0], complex(summand[1].real)] for summand in self.pauli_list
        ]
        return True

    @pydantic.validator("pauli_list", each_item=True)
    def validate_pauli_monomials(cls, monomial):
        parsed_monomial = PauliMonomial(string=monomial[0], coeff=monomial[1])
        return [parsed_monomial.string, parsed_monomial.coeff]

    @pydantic.validator("pauli_list")
    def validate_pauli_list(cls, pauli_list):
        if not all_equal(len(summand[0]) for summand in pauli_list):
            raise ValueError("Pauli strings have incompatible lengths.")
        return pauli_list


@pydantic.dataclasses.dataclass
class PauliMonomial:
    string: pydanticPauliMonomialStr
    coeff: Complex


class OperatorStatus(str, enum.Enum):
    SUCCESS = "success"
    ERROR = "error"


class OperatorResult(pydantic.BaseModel):
    status: OperatorStatus
    details: PauliOperator



from typing import Optional, Tuple, Union

import pydantic
from qiskit import circuit as qiskit_circuit, qasm as qiskit_qasm

from classiq_interface.generator.functions.register import Register
from classiq_interface.helpers.custom_pydantic_types import pydanticNonEmptyString

RegistersStrictType = Tuple[Register, ...]
RegistersType = Union[Register, RegistersStrictType]


def to_tuple(reg_type: RegistersType) -> RegistersStrictType:
    if isinstance(reg_type, Register):
        return (reg_type,)
    else:
        return reg_type


class FunctionImplementation(pydantic.BaseModel):
    """
    Facilitates the creation of a user-defined custom function implementation
    """

    class Config:
        validate_all = True

    name: Optional[pydanticNonEmptyString] = pydantic.Field(
        default=None,
        description="The name of the custom implementation",
    )

    num_qubits_in_serialized_circuit: Optional[pydantic.PositiveInt] = pydantic.Field(
        default=None,
        alias="_num_qubits_in_serialized_circuit",
        description="This is a private attribute",
    )

    serialized_circuit: pydanticNonEmptyString = pydantic.Field(
        description="The QASM code of the custom function implementation",
    )

    input_registers: RegistersType = pydantic.Field(
        default=tuple(),
        description="A tuple of input registers to the custom implementation",
    )

    output_registers: RegistersType = pydantic.Field(
        description="A tuple of output registers of the custom implementation",
    )

    zero_input_registers: RegistersType = pydantic.Field(
        default=tuple(),
        description="A tuple of zero input registers to the custom implementation",
    )

    auxiliary_registers: RegistersType = pydantic.Field(
        default=tuple(),
        description="A tuple of auxiliary registers to the custom implementation",
    )

    @pydantic.validator(
        "input_registers",
        "output_registers",
        "zero_input_registers",
        "auxiliary_registers",
        pre=True,
    )
    def validate_all_registers_are_tuples(
        cls,
        registers: RegistersType,
    ):
        if isinstance(registers, Register):
            return (registers,)
        return registers

    @pydantic.validator("serialized_circuit")
    def validate_serialized_circuit_and_num_qubits_in_serialized_circuit(
        cls, serialized_circuit, values
    ):
        try:
            qc = qiskit_circuit.QuantumCircuit.from_qasm_str(serialized_circuit)
        except qiskit_qasm.exceptions.QasmError:  # The qiskit error is often extremely uninformative
            raise ValueError("The QASM string is not a valid quantum circuit.")
        values["num_qubits_in_serialized_circuit"] = qc.num_qubits
        return serialized_circuit

    @pydantic.root_validator(skip_on_failure=True)
    def validate_ranges_of_all_registers(cls, values):
        input_registers: RegistersStrictType = values.get("input_registers")
        output_registers: RegistersStrictType = values.get("output_registers")
        zero_input_registers: RegistersStrictType = values.get("zero_input_registers")
        auxiliary_registers: RegistersStrictType = values.get("auxiliary_registers")
        all_qubits = set(range(values.get("num_qubits_in_serialized_circuit")))

        all_output_registers = output_registers + auxiliary_registers
        output_qubits = set().union(
            *[set(register.qubits) for register in all_output_registers]
        )
        if len(output_qubits) != sum(
            register.width for register in all_output_registers
        ):
            raise ValueError(
                "The output registers of a custom function must not overlap."
            )
        if output_qubits.difference(all_qubits):
            raise ValueError(
                "The number of qubits in the quantum circuit of the implementation must contain its output registers."
            )

        all_input_registers = (
            input_registers + zero_input_registers + auxiliary_registers
        )
        input_qubits = set().union(
            *[set(register.qubits) for register in all_input_registers]
        )
        if len(input_qubits) != sum(register.width for register in all_input_registers):
            raise ValueError(
                "The input registers of a custom function must not overlap."
            )
        if input_qubits != all_qubits:
            raise ValueError(
                "The number of qubits in the quantum circuit of the implementation must match its input registers."
            )

        return values

    @pydantic.validator("output_registers")
    def validate_output_registers(cls, output_registers: RegistersStrictType, values):
        if not output_registers:
            raise ValueError("The outputs of a custom function must be non-empty.")
        return output_registers

    @pydantic.validator("input_registers", "output_registers")
    def validate_io_registers_are_distinct(
        cls,
        io_registers: RegistersStrictType,
    ):
        if len(io_registers) != len({io_register.name for io_register in io_registers}):
            raise ValueError("The names of IO registers must be distinct.")
        return io_registers



from typing import Optional

import pydantic

from classiq_interface.chemistry.operator import PauliOperator
from classiq_interface.generator import function_params


class ExponentiationConstraints(pydantic.BaseModel):
    max_depth: Optional[pydantic.PositiveInt] = pydantic.Field(
        default=None, description="Maximum depth of the exponentiation circuit."
    )
    max_error: Optional[pydantic.PositiveFloat] = pydantic.Field(
        default=None,
        description="Maximum approximation error of the exponentiation circuit.",
    )


class Exponentiation(function_params.FunctionParams):
    """
    Exponantiation of a Hermitian Pauli sum operator.
    """

    pauli_operator: PauliOperator = pydantic.Field(
        description="A weighted sum of Pauli strings."
    )
    evolution_coefficient: float = pydantic.Field(
        default=1.0, description="A global coeffient multiplying the operator."
    )
    constraints: ExponentiationConstraints = pydantic.Field(
        default_factory=ExponentiationConstraints,
        description="Constraints for the exponentiation.",
    )
    _input_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_INPUT_NAME])
    _output_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_OUTPUT_NAME])

    @pydantic.validator("pauli_operator")
    def validate_is_hermitian(cls, pauli_operator: PauliOperator):
        if not pauli_operator.to_hermitian():
            raise ValueError("Cefficients of the Hamiltonian must be real numbers")
        return pauli_operator



import base64
import io
from datetime import datetime, time
from typing import List, Optional, Tuple

import matplotlib.pyplot as plt
import pydantic
from matplotlib.ticker import MaxNLocator
from PIL import Image
from tabulate import tabulate

from classiq_interface.helpers.custom_pydantic_types import pydanticProbabilityFloat

Solution = Tuple[int, ...]


class SolverResult(pydantic.BaseModel):
    best_cost: float
    # TODO: add time units (like seconds)
    time: time
    solution: Solution


class SolutionData(pydantic.BaseModel):
    solution: Solution
    repetitions: Optional[pydantic.PositiveInt]
    probability: Optional[pydanticProbabilityFloat]
    cost: float


class VQEIntermediateData(pydantic.BaseModel):
    utc_time: datetime = pydantic.Field(description="Time when the iteration finished")
    iteration_number: pydantic.PositiveInt = pydantic.Field(
        description="The iteration's number (evaluation count)"
    )
    parameters: List[float] = pydantic.Field(
        description="The optimizer parameters for the variational form"
    )
    mean_all_solutions: Optional[float] = pydantic.Field(
        default=None, description="The mean score of all solutions in this iteration"
    )
    solutions: List[SolutionData] = pydantic.Field(
        description="Solutions found in this iteration, their score and"
        "number of repetitions"
    )


_MAX_BIN = 50


class VQESolverResult(SolverResult):
    energy: Optional[float]
    solution_distribution: List[SolutionData] = []
    intermediate_results: List[VQEIntermediateData] = []
    optimal_parameters: List[float] = []
    convergence_graph_str: Optional[str]

    def show_convergence_graph(self) -> None:
        self.convergence_graph.show()

    @property
    def convergence_graph(self):
        return Image.open(io.BytesIO(base64.b64decode(self.convergence_graph_str)))

    def optimal_parameters_graph(self, num_lines: int = 2):
        layers = list(range(len(self.optimal_parameters) // num_lines))

        if num_lines == 2:
            legends = [r"$\gamma$", r"$\beta$"]
        else:
            legends = [f"line{i}" for i in range(num_lines)]

        for i in range(num_lines):
            plt.plot(
                layers,
                self.optimal_parameters[i::num_lines],
                marker="o",
                linestyle="--",
            )
        plt.xlabel("repetition")
        plt.ylabel("value")
        plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
        plt.legend(legends)

    def __str__(self) -> str:
        return "\n".join(self.formatted())

    def formatted(self) -> List[str]:
        lines = []
        lines.append("=== OPTIMAL SOLUTION ===")
        lines.append(
            tabulate(
                [[tuple(self.solution), self.best_cost]], headers=["solution", "cost"]
            )
        )

        lines.append("=== SOLUTION DISTRIBUTION ===")
        solution_distribution_table = [
            [solution_data.solution, solution_data.cost, solution_data.probability]
            for solution_data in self.solution_distribution
        ]
        lines.append(
            tabulate(
                solution_distribution_table, headers=["solution", "cost", "probability"]
            )
        )
        lines.append("=== OPTIMAL_PARAMETERS ===")
        lines.append(str(self.optimal_parameters))
        lines.append("=== TIME ===")
        lines.append(str(self.time))
        return lines

    def histogram(self):
        repetitions = [
            solution_data.repetitions for solution_data in self.solution_distribution
        ]
        costs = [solution_data.cost for solution_data in self.solution_distribution]

        bins = min(len(set(costs)), _MAX_BIN)
        eps = (max(costs) - min(costs)) / bins / 2
        hist_range = (min(costs) - eps, max(costs) + eps)
        plt.hist(
            x=costs, bins=bins, density=True, weights=repetitions, range=hist_range
        )

        plt.ylabel("Probability")
        plt.xlabel("value")



from typing import Set, Type

from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.generator.standard_gates.controlled_standard_gates import (
    C3XGate,
    C4XGate,
    CCXGate,
    CHGate,
    CPhaseGate,
    CRXGate,
    CRYGate,
    CRZGate,
    CSXGate,
    CXGate,
    CYGate,
    CZGate,
    MCPhaseGate,
)
from classiq_interface.generator.standard_gates.standard_angled_gates import (
    PhaseGate,
    RGate,
    RXGate,
    RXXGate,
    RYGate,
    RYYGate,
    RZGate,
    RZZGate,
)
from classiq_interface.generator.standard_gates.standard_gates import (
    HGate,
    IGate,
    SdgGate,
    SGate,
    SwapGate,
    SXdgGate,
    SXGate,
    TdgGate,
    TGate,
    XGate,
    YGate,
    ZGate,
    iSwapGate,
)

_standard_function_param_list: Set[Type[FunctionParams]] = {
    XGate,
    YGate,
    ZGate,
    HGate,
    IGate,
    SGate,
    SdgGate,
    SXGate,
    SXdgGate,
    TGate,
    TdgGate,
    iSwapGate,
    SwapGate,
    RXGate,
    RXXGate,
    RYGate,
    RYYGate,
    RZGate,
    RZZGate,
    RGate,
    PhaseGate,
    CXGate,
    CCXGate,
    C3XGate,
    C4XGate,
    CYGate,
    CZGate,
    CHGate,
    CRXGate,
    CRYGate,
    CRZGate,
    CSXGate,
    CPhaseGate,
    MCPhaseGate,
}


def get_qiskit_standard_function_param_list() -> Set[Type[FunctionParams]]:
    return _standard_function_param_list



from typing import List, Tuple

import numpy as np
import pydantic
from scipy.stats import norm

from classiq_interface.generator import function_params, linear_pauli_rotations
from classiq_interface.helpers.custom_pydantic_types import (
    pydanticNonOneProbabilityFloat,
    pydanticProbabilityFloat,
)

STATE_TARGET = linear_pauli_rotations.STATE_TARGET
RHOS_PZEROS_LENGTH_ERROR_MSG = "rhos and p_zeros must have the same length"


class LinearGCI(function_params.FunctionParams):
    """
    A circuit composed of a series of Y rotations to perform a linear approximation to the Gaussian Conditional
    Independence model.
    The model consists of a Bernoulli probability, which itself is a function of a latent random variable x:
    p(x) = F( (F_inv(p_zero) + x*sqrt(rho))/sqrt(1-rho) )
    with F being the Gaussian CDF, p_zero is p(x=0) when rho=0, and rho reflects the sensitivity.

    The circuit takes a state register |x> and zero-input target qubits.
    Each target qubit undergoes the following transformation:
    |x>|0> -> cos((slope*x + offset)/2)|x>|0> + sin((slope*x + offset)/2)|x>|1>
    Where the slope and the offset are determined by the linear approximation of the inverse sine of p(x).
    """

    _input_names = pydantic.PrivateAttr(default=STATE_TARGET)
    _output_names = pydantic.PrivateAttr(default=STATE_TARGET)

    num_state_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of input qubits"
    )
    truncation_value: float = pydantic.Field(
        description="The truncation value of the latent normal distribution"
    )
    p_zeros: List[pydanticProbabilityFloat] = pydantic.Field(
        description="The probability when the latent normal variable equals zero"
    )
    rhos: List[pydanticNonOneProbabilityFloat] = pydantic.Field(
        "The sensitivity of the probability to changes in the latent normal variable values"
    )

    @pydantic.validator("rhos")
    def validate_rhos(cls, rhos, values):
        p_zeros = values.get("p_zeros")
        if len(p_zeros) != len(rhos):
            raise ValueError(RHOS_PZEROS_LENGTH_ERROR_MSG)
        return rhos

    @property
    def linear_pauli_rotations(self) -> linear_pauli_rotations.LinearPauliRotations:
        num_state_qubits = self.num_state_qubits
        truncation_value = self.truncation_value
        rhos = self.rhos
        p_zeros = self.p_zeros
        offsets, slopes = get_gci_values(
            num_state_qubits, truncation_value, rhos, p_zeros
        )
        num_target_qubits = len(rhos)
        bases = ["Y"] * num_target_qubits

        return linear_pauli_rotations.LinearPauliRotations(
            num_state_qubits=num_state_qubits,
            bases=bases,
            offsets=offsets,
            slopes=slopes,
        )


def get_gci_values(
    qubit_count_state: int,
    truncation_value: float,
    rhos: List[float],
    p_zeros: List[float],
) -> Tuple[List[float], List[float]]:
    qubit_count_target = len(rhos)

    # Qiskit's code below

    # get normal (inverse) CDF and pdf
    def F(x: float) -> float:
        return norm.cdf(x)

    def F_inv(x: float) -> float:
        return norm.ppf(x)

    def f(x: float) -> float:
        return norm.pdf(x)

    # create linear rotations for conditional defaults
    slopes = []
    offsets = []
    for k in range(qubit_count_target):
        psi = F_inv(p_zeros[k]) / np.sqrt(1 - rhos[k])

        # compute slope / offset
        slope = -np.sqrt(rhos[k]) / np.sqrt(1 - rhos[k])
        slope *= f(psi) / np.sqrt(1 - F(psi)) / np.sqrt(F(psi))
        offset = 2 * np.arcsin(np.sqrt(F(psi)))

        # adjust for integer to normal range mapping
        offset += slope * (-truncation_value)
        slope *= 2 * truncation_value / (2 ** qubit_count_state - 1)

        offsets.append(offset)
        slopes.append(slope)

    return offsets, slopes



from typing import TYPE_CHECKING, List, Union

import pydantic

from classiq_interface.generator import function_params

STATE = "state"
TARGET = "target"
STATE_TARGET = [STATE, TARGET]

LENGTH_ERROR_MESSAGE = "Slopes, offsets and basis values must have the same length"

if TYPE_CHECKING:
    pydanticPauliBasisStr = str
else:
    pydanticPauliBasisStr = pydantic.constr(
        to_lower=True,
        regex=r"[xyz]",
    )


class LinearPauliRotations(function_params.FunctionParams):
    """
    Perform independent linear rotations on target qubits, each controlled by an identical
    n-qubit state register |x>.

    Each target qubit, indexed with k and denoted by q_k, undergoes the following transformation:
    |x>|q_k> -> |x> * [cos(theta(x,k)/2) + i*sin(theta(x,k)/2)*sigma]|q_k>
    with sigma being 'X', 'Y' or 'Z' Pauli matrix, and the angle is a linear function of the state,
    theta(x,k)/2 = (slope(k)*x + offset(k))/2.

    For example, a 'Y' rotation on one target qubit will result in a circuit implementing the following logic:
    |x>|0> -> cos((slope*x + offset)/2)|x>|0> + sin((slope*x + offset)/2)|x>|1>

            q_0: ─────────────────────────■───────── ... ──────────────────────
                                          │
                                          .
                                          │
        q_(n-1): ─────────────────────────┼───────── ... ───────────■──────────
                  ┌────────────┐  ┌───────┴───────┐       ┌─────────┴─────────┐
         target: ─┤ RY(offset) ├──┤ RY(2^0 slope) ├  ...  ┤ RY(2^(n-1) slope) ├
                  └────────────┘  └───────────────┘       └───────────────────┘
    """

    _input_names = pydantic.PrivateAttr(default=STATE_TARGET)
    _output_names = pydantic.PrivateAttr(default=STATE_TARGET)

    num_state_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of input qubits"
    )
    bases: Union[pydanticPauliBasisStr, List[pydanticPauliBasisStr]] = pydantic.Field(
        description="The types of Pauli rotations ('X', 'Y', 'Z')."
    )
    slopes: Union[float, List[float]] = pydantic.Field(
        description="The slopes of the controlled rotations."
    )
    offsets: Union[float, List[float]] = pydantic.Field(
        description="The offsets of the controlled rotations."
    )

    @pydantic.validator("bases")
    def validate_bases(cls, v):
        if not isinstance(v, list):
            v = [v]
        return v

    @pydantic.validator("slopes")
    def validate_slopes(cls, v):
        if not isinstance(v, list):
            v = [v]
        return v

    @pydantic.validator("offsets")
    def validate_offsets(cls, v):
        if not isinstance(v, list):
            v = [v]
        return v

    @pydantic.root_validator()
    def validate_lists(cls, values):
        offsets = values.get("offsets")
        bases = values.get("bases")
        slopes = values.get("slopes")
        if len(slopes) == len(offsets) and len(offsets) == len(bases):
            return values
        raise ValueError(LENGTH_ERROR_MESSAGE)



from typing import List, Literal, Optional, Union

import pydantic

from classiq_interface.generator import finance, grover_operator


class FinanceModelMetadata(pydantic.BaseModel):
    metadata_type: Literal["finance_model"] = "finance_model"
    num_model_qubits: int
    distribution_range: List[float]


class FunctionMetadata(pydantic.BaseModel):
    metadata_type: Literal["function"] = "function"
    name: str
    parent: Optional[str]
    children: List[str]


class GroverMetadata(grover_operator.GroverOperator):
    metadata_type: Literal["grover"] = "grover"


class FinanceMetadata(finance.Finance):
    metadata_type: Literal["finance"] = "finance"


MetadataUnion = Union[FinanceModelMetadata, GroverMetadata, FinanceMetadata]


class GenerationMetadata(pydantic.BaseModel):
    # Ideally, we would use a "__root__" attribute, but the typescript transpilation
    # does weird things when we use it.
    metadata: MetadataUnion



import enum
from typing import List, Optional, Union

import pydantic

from classiq_interface.backend.backend_preferences import (
    BackendPreferencesTypes,
    backend_preferences_field,
)
from classiq_interface.backend.quantum_backend_providers import IBMQBackendNames
from classiq_interface.executor.error_mitigation import ErrorMitigationMethod
from classiq_interface.generator.noise_properties import NoiseProperties
from classiq_interface.generator.preferences.randomness import create_random_seed
from classiq_interface.helpers.custom_pydantic_types import (
    pydanticAlphaParamCVAR,
    pydanticProbabilityFloat,
)


class AmplitudeEstimation(pydantic.BaseModel):
    alpha: float = pydantic.Field(
        default=0.05, description="Confidence level of the AE algorithm"
    )
    epsilon: float = pydantic.Field(
        default=0.01, description="precision for estimation target `a`"
    )
    binary_search_threshold: Optional[pydanticProbabilityFloat] = pydantic.Field(
        description="The required probability on the tail of the distribution (1 - percentile)"
    )

    objective_qubits: List[int] = pydantic.Field(
        default=[],
        description="list specifying on which qubits to perform the amplitude estimation",
    )


class AmplitudeAmplification(pydantic.BaseModel):
    iterations: Optional[Union[List[int], int]] = pydantic.Field(
        default=None,
        description="Number or list of numbers of iteration to use",
    )
    growth_rate: Optional[float] = pydantic.Field(
        default=None,
        description="Number of iteration used is set to round(growth_rate**iterations)",
    )
    sample_from_iterations: bool = pydantic.Field(
        default=False,
        description="If True, number of iterations used is picked randomly from "
        "[1, iteration] range",
    )


class CostType(str, enum.Enum):
    MIN = "MIN"
    AVERAGE = "AVERAGE"
    CVAR = "CVAR"


class OptimizerType(str, enum.Enum):
    COBYLA = "COBYLA"
    SPSA = "SPSA"
    L_BFGS_B = "L_BFGS_B"
    NELDER_MEAD = "NELDER_MEAD"
    ADAM = "ADAM"


class OptimizerPreferences(pydantic.BaseModel):
    type: OptimizerType = pydantic.Field(
        default=OptimizerType.COBYLA, description="Classical optimization algorithm."
    )
    num_shots: pydantic.PositiveInt = pydantic.Field(
        default=1000, description="Number of repetitions of the quantum ansatz."
    )
    cost_type: CostType = pydantic.Field(
        default=CostType.CVAR,
        description="Summarizing method of the measured bit strings",
    )
    alpha_cvar: pydanticAlphaParamCVAR = pydantic.Field(
        default=None, description="Parameter for the CVAR summarizing method"
    )
    max_iteration: pydantic.PositiveInt = pydantic.Field(
        default=100, description="Maximal number of optimizer iterations"
    )
    tolerance: pydantic.PositiveFloat = pydantic.Field(
        default=None, description="Final accuracy in the optimization"
    )
    step_size: pydantic.PositiveFloat = pydantic.Field(
        default=None,
        description="step size for numerically " "calculating the gradient",
    )
    random_seed: int = pydantic.Field(
        default_factory=create_random_seed,
        description="The random seed used for the generation",
    )
    initial_point: Optional[List[float]] = pydantic.Field(
        default=None,
        description="Initial values for the ansatz parameters",
    )

    @pydantic.validator("alpha_cvar", pre=True, always=True)
    def check_alpha_cvar(cls, alpha_cvar, values):
        cost_type = values.get("cost_type")
        if alpha_cvar is not None and cost_type != CostType.CVAR:
            raise ValueError("Use CVAR params only for CostType.CVAR.")

        if alpha_cvar is None and cost_type == CostType.CVAR:
            alpha_cvar = 0.2

        return alpha_cvar

    @pydantic.validator("tolerance", pre=True, always=True)
    def check_tolerance(cls, tolerance, values):
        optimizer_type = values.get("type")
        if tolerance is not None and optimizer_type == OptimizerType.SPSA:
            raise ValueError("No tolerance param for SPSA optimizer")

        if tolerance is None and optimizer_type != OptimizerType.SPSA:
            tolerance = 0.001

        return tolerance

    @pydantic.validator("step_size", pre=True, always=True)
    def check_step_size(cls, step_size, values):
        optimizer_type = values.get("type")
        if step_size is not None and optimizer_type not in (
            OptimizerType.L_BFGS_B,
            OptimizerType.ADAM,
        ):
            raise ValueError("Use step_size only for L_BFGS_B or ADAM optimizers.")

        if step_size is None and optimizer_type in (
            OptimizerType.L_BFGS_B,
            OptimizerType.ADAM,
        ):
            step_size = 0.05

        return step_size


class VQEOptimization(pydantic.BaseModel):
    optimizer_preferences: OptimizerPreferences = pydantic.Field(
        default_factory=OptimizerPreferences,
        description="preferences for the VQE execution",
    )
    is_maximization: bool = pydantic.Field(
        default=False,
        description="Whether the VQE goal is to maximize",
    )


class ExecutionPreferences(pydantic.BaseModel):
    num_shots: pydantic.PositiveInt = 100
    timeout_sec: Optional[pydantic.PositiveInt] = pydantic.Field(
        default=None,
        description="If set, limits the execution runtime. Value is in seconds. "
        "Not supported on all platforms.",
    )
    amplitude_estimation: Optional[AmplitudeEstimation]
    amplitude_amplification: Optional[AmplitudeAmplification]
    vqe_optimization: Optional[VQEOptimization]
    error_mitigation_method: Optional[ErrorMitigationMethod] = pydantic.Field(
        default=None,
        description="Error mitigation method. Currently supports complete and tensored "
        "measurement calibration.",
    )
    noise_properties: Optional[NoiseProperties] = pydantic.Field(
        default=None, description="Properties of the noise in the circuit"
    )
    random_seed: int = pydantic.Field(
        default_factory=create_random_seed,
        description="The random seed used for the generation",
    )
    backend_preferences: BackendPreferencesTypes = backend_preferences_field(
        backend_name=IBMQBackendNames.IBMQ_AER_SIMULATOR_STATEVECTOR
    )



import enum
from typing import Dict, List, Optional, Union

import pydantic


class RotationBlocksType(str, enum.Enum):
    rx = "rx"
    ry = "ry"
    rz = "rz"


class EntanglementBlocksType(str, enum.Enum):
    cx = "cx"
    cy = "cy"
    cz = "cz"


class EntanglementStructureType(str, enum.Enum):
    linear = "linear"
    full = "full"
    circular = "circular"
    sca = "sca"


class CustomAnsatzArgs(pydantic.BaseModel):
    num_qubits: int


class SeparateU3Args(CustomAnsatzArgs):
    pass


class HypercubeArgs(CustomAnsatzArgs):
    layer_count: int = 2


class EntanglingLayersArgs(CustomAnsatzArgs):
    layer_count: int = 2


class RandomArgs(CustomAnsatzArgs):
    gate_count: int = 100
    gate_probabilities: Dict[str, float] = {"cx": 0.5, "u": 0.5}
    random_seed: Optional[int] = None


class RandomTwoQubitGatesArgs(CustomAnsatzArgs):
    random_two_qubit_gate_count_factor: float = 1.0
    random_seed: Optional[int] = None


class TwoLocalArgs(CustomAnsatzArgs):
    rotation_blocks: Optional[
        Union[RotationBlocksType, List[RotationBlocksType]]
    ] = RotationBlocksType.ry
    entanglement_blocks: Optional[
        Union[EntanglementBlocksType, List[EntanglementBlocksType]]
    ] = EntanglementBlocksType.cx
    entanglement: EntanglementStructureType = EntanglementStructureType.full
    reps: int = 3



import itertools
from typing import Union

import networkx as nx
import pyomo.core as pyo


def tsp_digraph(
    graph: nx.DiGraph, source: Union[int, str], sink: Union[int, str], nonedge: int
) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()

    assert source in graph.nodes(), "source node is out of range"
    assert sink in graph.nodes(), "sink node is out of range"

    points = list(graph.nodes() - {source, sink})
    steps = range(1, graph.number_of_nodes() - 1)
    idxs = list(itertools.product(points, steps))
    model.x = pyo.Var(
        idxs, domain=pyo.Binary
    )  # x[i, j] = 1 indicates that point i is visited at step j

    @model.Constraint(points)
    def each_step_visits_one_point_rule(model, ii):
        return sum(model.x[ii, jj] for jj in steps) == 1

    @model.Constraint(steps)
    def each_point_visited_once_rule(model, jj):
        return sum(model.x[ii, jj] for ii in points) == 1

    def is_travel_between_2_points(point1: int, point2: int):
        if point1 == source:
            if point2 == sink:
                return 0
            return model.x[point2, steps[0]]

        if point1 == sink:
            return 0

        if point2 == source:
            return 0

        if point2 == sink:
            return model.x[point1, steps[-1]]

        return sum(model.x[point1, kk] * model.x[point2, kk + 1] for kk in steps[:-1])

    model.cost = pyo.Objective(
        expr=sum(
            graph.get_edge_data(point1, point2, default={"weight": nonedge})["weight"]
            * is_travel_between_2_points(point1, point2)
            for point1, point2 in itertools.product(graph.nodes, repeat=2)
            if point1 != point2
        )
    )
    return model



from typing import List, Union

import pydantic

from classiq_interface.generator import function_params

# TODO mutable defaults might be problematic @shahakL


class HardwareEfficientAnsatz(function_params.FunctionParams):
    _input_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_INPUT_NAME])
    _output_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_OUTPUT_NAME])

    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="Number of qubits in the ansatz"
    )
    connectivity_map: List[List[pydantic.NonNegativeInt]] = pydantic.Field(
        description="Hardware's connectivity map, in the form [ [x0, x1], [x1, x2],...]"
    )
    reps: pydantic.PositiveInt = pydantic.Field(
        default=1, description="Number of layers in the Ansatz"
    )

    one_qubit_gates: Union[str, List[str]] = pydantic.Field(
        default=["x", "ry"],
        description='List of gates for the one qubit gates layer, e.g. ["x", "ry"]',
    )
    two_qubit_gates: Union[str, List[str]] = pydantic.Field(
        default=["cx"],
        description='List of gates for the two qubit gates entangling layer, e.g. ["cx", "cry"]',
    )

    @pydantic.validator("connectivity_map")
    def validate_connectivity_map(cls, connectivity_map, values):
        num_qubits = values.get("num_qubits")
        if num_qubits is None:
            return
        for connection in connectivity_map:
            if len(connection) != 2:
                raise ValueError(
                    f"Connectivity {connection} is invalid, must connect exactly 2 qubits"
                )
            for index in connection:
                if index >= num_qubits:
                    raise ValueError(
                        f"Provided index {index} is bigger than num_qubits-1"
                    )
        return connectivity_map



import networkx as nx
import pyomo.core as pyo


def mis(graph: nx.Graph) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    model.x = pyo.Var(graph.nodes, domain=pyo.Binary)

    @model.Constraint(graph.edges)
    def independent_rule(model, node1, node2):
        return model.x[node1] + model.x[node2] <= 1

    model.cost = pyo.Objective(expr=sum(model.x.values()), sense=pyo.maximize)

    return model



from typing import Any, Dict, Literal, Union

import pydantic

from classiq_interface.backend.backend_preferences import IonqBackendPreferences
from classiq_interface.executor.execution_preferences import ExecutionPreferences
from classiq_interface.executor.hamiltonian_minimization_problem import (
    HamiltonianMinimizationProblem,
)
from classiq_interface.executor.quantum_program import (
    QuantumInstructionSet,
    QuantumProgram,
)
from classiq_interface.generator.generation_metadata import GenerationMetadata


class GenerationMetadataExecution(GenerationMetadata):
    execution_type: Literal["generation_metadata"] = "generation_metadata"


class QuantumProgramExecution(QuantumProgram):
    execution_type: Literal["quantum_program"] = "quantum_program"


class HamiltonianMinimizationProblemExecution(HamiltonianMinimizationProblem):
    execution_type: Literal[
        "hamiltonian_minimization_problem"
    ] = "hamiltonian_minimization_problem"


class ExecutionRequest(pydantic.BaseModel):
    execution_payload: Union[
        GenerationMetadataExecution,
        QuantumProgramExecution,
        HamiltonianMinimizationProblemExecution,
    ]
    preferences: ExecutionPreferences = pydantic.Field(
        default_factory=ExecutionPreferences,
        description="preferences for the execution",
    )

    @pydantic.validator("preferences")
    def validate_ionq_backend(
        cls, preferences: ExecutionPreferences, values: Dict[str, Any]
    ):
        quantum_program = values.get("execution_payload")
        if isinstance(quantum_program, QuantumProgram):
            if quantum_program.syntax == QuantumInstructionSet.IONQ:
                raise ValueError("Can only execute IonQ code on IonQ backend.")
        elif isinstance(preferences.backend_preferences, IonqBackendPreferences):
            raise ValueError("IonQ backend supports only execution of QuantumPrograms")
        return preferences



from typing import Literal, Union

import pydantic

from classiq_interface.chemistry.ground_state_problem import GroundStateProblem
from classiq_interface.generator import function_params


class HartreeFock(function_params.FunctionParams):
    gs_problem: Union[Literal["ground_state_problem"], GroundStateProblem]
    _input_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_INPUT_NAME])
    _output_names = pydantic.PrivateAttr(default=[function_params.DEFAULT_OUTPUT_NAME])

    @pydantic.validator("gs_problem")
    def validate_gs_problem(cls, gs_problem):
        if not isinstance(gs_problem, GroundStateProblem):
            raise ValueError("ground state problem must be of type GroundStateProblem")
        return gs_problem



import itertools

import numpy as np
import pyomo.core as pyo


def tsp(distance_matrix: np.ndarray) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()

    assert (
        distance_matrix.shape[0] == distance_matrix.shape[1]
    ), "distance_matrix is not square"

    num_points = distance_matrix.shape[0]

    points = range(num_points)
    idxs = list(itertools.product(points, repeat=2))
    model.x = pyo.Var(
        idxs, domain=pyo.Binary
    )  # x[i, j] = 1 indicates that point i is visited at step j

    @model.Constraint(points)
    def each_step_visits_one_point_rule(model, ii):
        return sum(model.x[ii, jj] for jj in range(num_points)) == 1

    @model.Constraint(points)
    def each_point_visited_once_rule(model, jj):
        return sum(model.x[ii, jj] for ii in range(num_points)) == 1

    def is_travel_between_2_points(point1: int, point2: int):
        return sum(model.x[point1, kk] * model.x[point2, kk + 1] for kk in points[:-1])

    model.cost = pyo.Objective(
        expr=sum(
            distance_matrix[point1, point2] * is_travel_between_2_points(point1, point2)
            for point1, point2 in idxs
        )
    )

    return model



import pyomo.environ as pyo


def is_maximization(model: pyo.ConcreteModel):
    objectives = list(model.component_objects(pyo.Objective))
    assert len(objectives) == 1, "supports only a single objective"
    return objectives[0].sense == pyo.maximize



from typing import Optional

import pydantic


class HardwareEfficientConstraints(pydantic.BaseModel):
    num_qubits: int
    num_two_qubit_gates: Optional[int] = None
    num_one_qubit_gates: Optional[int] = None
    max_depth: Optional[int] = None


class HardwareEfficient(pydantic.BaseModel):
    structure: str
    constraints: HardwareEfficientConstraints



from typing import Tuple, Union

import pydantic

from classiq_interface.finance.function_input import FinanceFunctionInput
from classiq_interface.finance.model_input import FinanceModelInput
from classiq_interface.generator import function_params


class Finance(function_params.FunctionParams):
    model: FinanceModelInput = pydantic.Field(description="Load a financial model")
    finance_function: Union[FinanceFunctionInput] = pydantic.Field(
        description="The finance function to solve the model"
    )

    _input_names = pydantic.PrivateAttr(default_factory=list)
    _output_names = pydantic.PrivateAttr(default=["out"])


class FinanceModels(function_params.FunctionParams):
    model: FinanceModelInput = pydantic.Field(description="Load a financial model")
    _input_names = pydantic.PrivateAttr(default=["in"])
    _output_names = pydantic.PrivateAttr(default=["out"])


class FinancePayoff(function_params.FunctionParams):
    finance_function: FinanceFunctionInput = pydantic.Field(
        description="The finance function to solve the model"
    )
    num_qubits: pydantic.PositiveInt
    distribution_range: Tuple[float, float]

    _input_names = pydantic.PrivateAttr(default=["in"])
    _output_names = pydantic.PrivateAttr(default=["out"])



from datetime import timedelta
from typing import Union

import pydantic

from classiq_interface.backend import pydantic_backend
from classiq_interface.backend.quantum_backend_providers import (
    EXACT_SIMULATORS,
    AWSBackendNames,
    AzureQuantumBackendNames,
    IBMQBackendNames,
    IonqBackendNames,
    ProviderTypeVendor,
    ProviderVendor,
)


class BackendPreferences(pydantic.BaseModel):
    backend_service_provider: str = pydantic.Field(
        description="Provider company for the requested backend."
    )
    backend_name: str = pydantic.Field(description="Name of the requested backend.")


AWS_DEFAULT_JOB_TIMEOUT_SECONDS = timedelta(minutes=5).total_seconds()


class AwsBackendPreferences(BackendPreferences):
    backend_service_provider: ProviderTypeVendor.AWS_BRAKET = ProviderVendor.AWS_BRAKET
    # Allow running any backend supported by the vendor
    backend_name: Union[AWSBackendNames, str]
    aws_access_key_id: pydantic_backend.pydanticAWSAccessKeyID = pydantic.Field(
        description="AWS Access Key ID"
    )
    aws_secret_access_key: pydantic_backend.pydanticAWSSecretAccessKey = pydantic.Field(
        description="AWS Secret Access Key"
    )
    region_name: pydantic_backend.pydanticRegionName = pydantic.Field(
        description="AWS Region name"
    )
    s3_bucket_name: pydantic_backend.pydanticS3BucketName = pydantic.Field(
        description="S3 Bucket Name"
    )
    s3_bucket_key: pydantic_backend.pydanticS3BucketKey = pydantic.Field(
        description="S3 Bucket Key"
    )
    job_timeout: pydantic_backend.pydanticExecutionTimeout = pydantic.Field(
        description="Timeout for Jobs sent for execution in seconds.",
        default=AWS_DEFAULT_JOB_TIMEOUT_SECONDS,
    )


class IBMBackendPreferences(BackendPreferences):
    backend_service_provider: ProviderTypeVendor.IBMQ = ProviderVendor.IBMQ
    backend_name: IBMQBackendNames


class AzureCredential(pydantic.BaseSettings):
    tenant_id: str = pydantic.Field(..., description="Azure Tenant ID")
    client_id: str = pydantic.Field(..., description="Azure Client ID")
    client_secret: str = pydantic.Field(..., description="Azure Client Secret")

    class Config:
        title = "Azure Service Principal Credential"
        env_prefix = "AZURE_"
        case_sensitive = False


class AzureBackendPreferences(BackendPreferences):
    backend_service_provider: ProviderTypeVendor.AZURE_QUANTUM = (
        ProviderVendor.AZURE_QUANTUM
    )
    # Allow running any backend supported by the vendor
    backend_name: Union[AzureQuantumBackendNames, str]

    resource_id: pydantic_backend.pydanticAzureResourceIDType = pydantic.Field(
        ...,
        description="Azure Resource ID (including Azure subscription ID, resource "
        "group and workspace)",
    )

    location: str = pydantic.Field(..., description="Azure Region")

    credential: AzureCredential = pydantic.Field(
        default_factory=AzureCredential,
        description="The service principal credential to access the quantum workspace",
    )


class IonqBackendPreferences(BackendPreferences):
    backend_service_provider: ProviderTypeVendor.IONQ = ProviderVendor.IONQ
    backend_name: IonqBackendNames = pydantic.Field(
        default=IonqBackendNames.SIMULATOR,
        description="IonQ backend for quantum programs execution.",
    )
    api_key: pydantic_backend.pydanticIonQApiKeyType = pydantic.Field(
        ..., description="IonQ API key"
    )


def is_exact_simulator(backend_preferences: BackendPreferences):
    return backend_preferences.backend_name in EXACT_SIMULATORS


def default_backend_preferences(
    backend_name: str = IBMQBackendNames.IBMQ_AER_SIMULATOR,
) -> BackendPreferences:
    return IBMBackendPreferences(backend_name=backend_name)


def backend_preferences_field(
    backend_name: str = IBMQBackendNames.IBMQ_AER_SIMULATOR,
) -> pydantic.Field:
    return pydantic.Field(
        default_factory=lambda: default_backend_preferences(backend_name),
        description="Preferences for the requested backend to run the quantum circuit.",
        # TODO add the following field once we switch to pydantic v1.9.0 (currently in pre-release)
        # discriminator="backend_service_provider",
    )


BackendPreferencesTypes = Union[
    AzureBackendPreferences,
    IBMBackendPreferences,
    AwsBackendPreferences,
    IonqBackendPreferences,
]



from enum import Enum
from typing import Any, Dict, Generic, List, Literal, Optional, Tuple, TypeVar, Union

import pydantic
from pydantic.generics import GenericModel

from classiq_interface.generator.arith.arithmetic import DEFAULT_OUT_NAME
from classiq_interface.generator.arith.fix_point_number import FixPointNumber
from classiq_interface.generator.arith.register_user_input import RegisterUserInput
from classiq_interface.generator.function_params import FunctionParams

DEFAULT_RIGHT_ARG_NAME = "right_arg"
DEFAULT_LEFT_ARG_NAME = "left_arg"
LeftDataT = TypeVar("LeftDataT")
RightDataT = TypeVar("RightDataT")
Numeric = (float, int)


class BinaryOpParams(GenericModel, FunctionParams, Generic[LeftDataT, RightDataT]):
    left_arg: LeftDataT
    right_arg: RightDataT
    output_size: Optional[pydantic.PositiveInt]
    output_name: str = DEFAULT_OUT_NAME

    @pydantic.validator("left_arg")
    def set_left_arg_name(cls, left_arg):
        if isinstance(left_arg, RegisterUserInput) and left_arg.name is None:
            left_arg.name = DEFAULT_LEFT_ARG_NAME
        return left_arg

    @pydantic.validator("right_arg")
    def set_right_arg_name(cls, right_arg):
        if isinstance(right_arg, RegisterUserInput) and right_arg.name is None:
            right_arg.name = DEFAULT_RIGHT_ARG_NAME
        return right_arg

    @pydantic.root_validator()
    def validate_one_is_register(cls, values):
        if isinstance(values.get("left_arg"), Numeric) and isinstance(
            values.get("right_arg"), Numeric
        ):
            raise ValueError("One argument must be a register")
        return values

    def create_io_enums(self):
        self._create_input_enum()
        self._create_output_enum()

    def _create_input_enum(self):
        input_name_dict = self._input_register_name_dict(
            [self.left_arg, self.right_arg]
        )
        assert input_name_dict, "At least one argument should be a register"
        self._input_enum = Enum("BinaryOpInputs", input_name_dict)

    def _create_output_enum(self):
        output_name_dict: Dict[str, str] = self._carried_inputs_name_dict()
        output_name_dict[self.output_name] = self.output_name
        self._output_enum = Enum("BinaryOpOutputs", output_name_dict)

    def _carried_inputs_name_dict(self) -> Dict[str, str]:
        return self._input_register_name_dict(list(self._carried_arguments()))

    def _carried_arguments(self) -> Tuple[Optional[LeftDataT], Optional[RightDataT]]:
        if getattr(self, "inplace", False):
            if isinstance(self.right_arg, RegisterUserInput):
                return self.left_arg, None
            else:
                return None, None
        return self.left_arg, self.right_arg

    @staticmethod
    def _input_register_name_dict(possible_register_args: List[Any]) -> Dict[str, str]:
        name_dict: Dict[str, str] = dict()
        for arg in possible_register_args:
            if isinstance(arg, RegisterUserInput) and arg.name:
                name_dict[arg.name] = arg.name
        return name_dict

    class Config:
        arbitrary_types_allowed = True


class BinaryOpWithIntInputs(
    BinaryOpParams[Union[int, RegisterUserInput], Union[int, RegisterUserInput]]
):
    @pydantic.root_validator()
    def validate_int_registers(cls, values):
        left_arg = values.get("left_arg")
        is_left_arg_float_register = (
            isinstance(left_arg, RegisterUserInput) and left_arg.fraction_places > 0
        )
        right_arg = values.get("right_arg")
        is_right_arg_float_register = (
            isinstance(right_arg, RegisterUserInput) and right_arg.fraction_places > 0
        )
        if is_left_arg_float_register or is_right_arg_float_register:
            raise ValueError("Boolean operation are defined only for integer")

        return values


class BinaryOpWithFloatInputs(
    BinaryOpParams[
        Union[float, FixPointNumber, RegisterUserInput],
        Union[float, FixPointNumber, RegisterUserInput],
    ]
):
    @pydantic.validator("left_arg", "right_arg")
    def convert_numeric_to_fix_point_number(cls, val):
        if isinstance(val, Numeric):
            val = FixPointNumber(float_value=val)
        return val


class BitwiseAnd(BinaryOpWithIntInputs):
    pass


class BitwiseOr(BinaryOpWithIntInputs):
    pass


class BitwiseXor(BinaryOpWithIntInputs):
    pass


class Adder(BinaryOpWithFloatInputs):
    inplace: bool = True


class Subtractor(BinaryOpWithFloatInputs):
    inplace: bool = True


class Multiplier(BinaryOpWithFloatInputs):
    pass


class Comparator(BinaryOpWithFloatInputs):
    output_size: Literal[1] = 1
    _include_equal: bool = pydantic.PrivateAttr(default=True)


class Equal(Comparator):
    pass


class NotEqual(Comparator):
    pass


class GreaterThan(Comparator):
    pass


class GreaterEqual(Comparator):
    pass


class LessThan(Comparator):
    pass


class LessEqual(Comparator):
    pass


class BinaryOpWithLeftRegRightInt(
    BinaryOpParams[RegisterUserInput, pydantic.PositiveInt]
):
    pass


class LShift(BinaryOpWithLeftRegRightInt):
    pass


class RShift(BinaryOpWithLeftRegRightInt):
    use_arithmetic_bit_shift: bool = True


class CyclicShift(BinaryOpParams[RegisterUserInput, int]):
    pass



from enum import Enum
from typing import Literal


class ProviderVendor(str, Enum):
    IBMQ = "IBMQ"
    AZURE_QUANTUM = "Azure Quantum"
    AWS_BRAKET = "AWS Braket"
    IONQ = "IonQ"


class ProviderTypeVendor:
    IBMQ = Literal[ProviderVendor.IBMQ]
    AZURE_QUANTUM = Literal[ProviderVendor.AZURE_QUANTUM]
    AWS_BRAKET = Literal[ProviderVendor.AWS_BRAKET]
    IONQ = Literal[ProviderVendor.IONQ]


class IonqBackendNames(str, Enum):
    SIMULATOR = "simulator"
    QPU = "qpu"


class AzureQuantumBackendNames(str, Enum):
    IONQ_SIMULATOR = "ionq.simulator"
    IONQ_QPU = "ionq.qpu"
    HONEYWELL_API_VALIDATOR1 = "honeywell.hqs-lt-s1-apival"
    HONEYWELL_API_VALIDATOR2 = "honeywell.hqs-lt-s2-apival"
    HONEYWELL_SIMULATOR1 = "honeywell.hqs-lt-s1-sim"
    HONEYWELL_SIMULATOR2 = "honeywell.hqs-lt-s2-sim"
    HONEYWELL_QPU1 = "honeywell.hqs-lt-s1"
    HONEYWELL_QPU2 = "honeywell.hqs-lt-s2"
    MICROSOFT_FULLSTATE_SIMULATOR = "microsoft.simulator.fullstate"


class AWSBackendNames(str, Enum):
    AWS_BRAKET_SV1 = "SV1"
    AWS_BRAKET_TN1 = "TN1"
    AWS_BRAKET_DM1 = "dm1"
    AWS_BRAKET_ASPEN_9 = "Aspen-9"
    AWS_BRAKET_ASPEN_10 = "Aspen-10"
    AWS_BRAKET_IONQ = "IonQ Device"


class IBMQBackendNames(str, Enum):
    IBMQ_AER_SIMULATOR = "aer_simulator"
    IBMQ_AER_SIMULATOR_STATEVECTOR = "aer_simulator_statevector"
    IBMQ_AER_SIMULATOR_DENSITY_MATRIX = "aer_simulator_density_matrix"
    IBMQ_AER_SIMULATOR_MATRIX_PRODUCT_STATE = "aer_simulator_matrix_product_state"


EXACT_SIMULATORS = [
    IonqBackendNames.SIMULATOR,
    AzureQuantumBackendNames.IONQ_SIMULATOR,
    AzureQuantumBackendNames.MICROSOFT_FULLSTATE_SIMULATOR,
    AWSBackendNames.AWS_BRAKET_SV1,
    AWSBackendNames.AWS_BRAKET_TN1,
    AWSBackendNames.AWS_BRAKET_DM1,
    *IBMQBackendNames,
]



import pydantic

from classiq_interface.helpers.custom_pydantic_types import pydanticLargerThanOneInteger


class SquareClusterEntanglerParameters(pydantic.BaseModel):
    num_of_qubits: pydanticLargerThanOneInteger
    schmidt_rank: pydantic.NonNegativeInt


class Open2DClusterEntanglerParameters(pydantic.BaseModel):
    qubit_count: pydanticLargerThanOneInteger
    schmidt_rank: pydantic.NonNegativeInt



from typing import Set

import pydantic

from classiq_interface.generator.function_params import FunctionParams


class CustomFunction(FunctionParams):
    """
    A user-defined custom function parameters object.
    """

    name: str = pydantic.Field(description="The name of a custom function")

    def generate_io_names(self, input_set: Set[str], output_set: Set[str]):
        self._input_names = list(input_set)
        self._output_names = list(output_set)



from enum import Enum
from typing import List, Optional, Set, Union

import pydantic

Match = List[List[int]]


class EntanglementAnalysisStatus(str, Enum):
    SUCCESS = "success"
    TIMEOUT = "timeout"
    ERROR = "error"


class EntanglementAnalysisResult(pydantic.BaseModel):
    status: EntanglementAnalysisStatus
    details: Union[pydantic.NonNegativeInt, str]


class AnalysisStatus(str, Enum):
    NONE = "none"
    SUCCESS = "success"
    CANCELLED = "cancelled"
    ERROR = "error"


class BasisGates(str, Enum):
    CX = "cx"
    CY = "cy"
    CZ = "cz"
    U = "u"
    U2 = "u2"
    P = "p"


class QuantumCircuitProperties(pydantic.BaseModel):
    depth: pydantic.NonNegativeInt = pydantic.Field(
        default=..., description="Circuit depth"
    )
    auxiliary_qubits: pydantic.NonNegativeInt = pydantic.Field(
        default=..., description="Number of Auxiliary qubits"
    )
    classical_bits: pydantic.NonNegativeInt = pydantic.Field(
        default=..., description="Number of classical bits"
    )
    gates_count: pydantic.NonNegativeInt = pydantic.Field(
        default=..., description="Total number of gates in the circuit"
    )
    multi_qubit_gates_count: pydantic.NonNegativeInt = pydantic.Field(
        default=..., description="Number of multi-qubit gates in circuit"
    )
    non_entangled_subcircuits_count: pydantic.NonNegativeInt = pydantic.Field(
        default=..., description="Number of non-entangled sub-circuit "
    )
    entanglement_upper_bound: EntanglementAnalysisResult = pydantic.Field(
        default=...,
        description="An upper bound to the entanglement (measured by the Schmidt rank width) of states that can"
        "generated by the circuit. None is returned if the entanglement analysis took too long to complete",
    )


class NativeQuantumCircuitProperties(QuantumCircuitProperties):
    native_gates: Set[BasisGates] = pydantic.Field(
        default=..., description="Native gates used for decomposition"
    )


class PatternRecognitionResult(pydantic.BaseModel):
    found_patterns: List[str]


class PatternMatchingResult(pydantic.BaseModel):
    found_patterns: List[str]


class Circuit(pydantic.BaseModel):
    closed_circuit_qasm: str
    image: str


class PatternAnalysis(pydantic.BaseModel):
    pattern_matching: Optional[PatternMatchingResult] = pydantic.Field(
        default=..., description="Pattern matching algorithm"
    )
    pattern_recognition: Optional[PatternRecognitionResult] = pydantic.Field(
        default=..., description="Find unknown patterns"
    )
    circuit: Circuit = pydantic.Field(
        default=..., description="Quantum circuit after pattern analysis"
    )


class Analysis(pydantic.BaseModel):
    input_properties: QuantumCircuitProperties = pydantic.Field(
        default=..., description="Input circuit properties"
    )
    native_properties: NativeQuantumCircuitProperties = pydantic.Field(
        default=..., description="Transpiled circuit properties"
    )
    pattern_analysis: Optional[PatternAnalysis] = pydantic.Field(
        default=None,
        description="Pattern analysis, including pattern matching and pattern recognition",
    )


class AnalysisResult(pydantic.BaseModel):
    status: AnalysisStatus
    details: Union[Analysis, str]



from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import Dict, List, Optional

import networkx as nx

from classiq_interface.generator.function_call import FunctionCall
from classiq_interface.helpers.custom_pydantic_types import pydanticNonEmptyString

IO_MULTI_USE_ERROR_MSG = "Input and output names can only be used once"
UNCONNECTED_WIRES_ERROR_MSG = "Wires connected only on one end"
UNCONNECTED_FLOW_IO_ERROR_MSG = (
    "Composite function inputs/outputs not connected to inner calls"
)
CYCLE_ERROR_MSG = "Inputs and outputs cannot form a cycle"


@dataclass
class Wire:
    start: Optional[pydanticNonEmptyString] = None
    end: Optional[pydanticNonEmptyString] = None


def _parse_call_inputs(
    function_call: FunctionCall, wires: Dict[str, Wire], flow_input_names: List[str]
) -> None:
    if not function_call.inputs:
        return

    for wire_name in function_call.inputs.values():
        if wire_name in flow_input_names:
            continue

        wire = wires[wire_name]

        if wire.end:
            raise ValueError(
                IO_MULTI_USE_ERROR_MSG
                + f". The name {wire_name} is used multiple times."
            )
        wire.end = function_call.name


def _parse_call_outputs(
    function_call: FunctionCall, wires: Dict[str, Wire], flow_output_names: List[str]
) -> None:
    if not function_call.outputs:
        return

    for wire_name in function_call.outputs.values():
        if wire_name in flow_output_names:
            continue

        wire = wires[wire_name]

        if wire.start:
            raise ValueError(
                IO_MULTI_USE_ERROR_MSG
                + f". The name {wire_name} is used multiple times."
            )
        wire.start = function_call.name


def _create_flow_graph(
    logic_flow: List[FunctionCall],
    flow_input_names: List[str],
    flow_output_names: List[str],
) -> nx.DiGraph:
    wires: Dict[str, Wire] = defaultdict(Wire)
    for function_call in logic_flow:
        _parse_call_inputs(
            function_call=function_call, wires=wires, flow_input_names=flow_input_names
        )
        _parse_call_outputs(
            function_call=function_call,
            wires=wires,
            flow_output_names=flow_output_names,
        )

    edges = [(wire.start, wire.end) for wire in wires.values()]

    graph = nx.DiGraph()

    graph.add_nodes_from(
        [(function_call.name, {"data": function_call}) for function_call in logic_flow]
    )
    graph.add_edges_from(edges)

    return graph


def _validate_io_names_match(
    logic_flow: List[FunctionCall],
    flow_input_names: List[str],
    flow_output_names: List[str],
) -> None:
    call_input_names = list(
        chain(
            *[
                function_call.inputs.values()
                for function_call in logic_flow
                if function_call.inputs
            ]
        )
    )
    call_output_names = list(
        chain(
            *[
                function_call.outputs.values()
                for function_call in logic_flow
                if function_call.outputs
            ]
        )
    )

    if sorted(call_input_names + flow_output_names) == sorted(
        call_output_names + flow_input_names
    ):
        return

    error_messages = list()
    unconnected_flow_ios = [
        name for name in flow_input_names if name not in call_input_names
    ] + [name for name in flow_output_names if name not in call_output_names]
    if unconnected_flow_ios:
        error_messages.append(
            f"{UNCONNECTED_FLOW_IO_ERROR_MSG}: {unconnected_flow_ios}"
        )

    unconnected_wires = [
        name
        for name in call_input_names
        if name not in call_output_names and name not in flow_input_names
    ] + [
        name
        for name in call_output_names
        if name not in call_input_names and name not in flow_output_names
    ]
    if unconnected_wires:
        error_messages.append(f"{UNCONNECTED_WIRES_ERROR_MSG}: {unconnected_wires}")

    raise ValueError("\n".join(error_messages))


def validate_flow_graph(
    logic_flow: List[FunctionCall],
    flow_input_names: List[str] = None,
    flow_output_names: List[str] = None,
) -> None:
    flow_input_names = list() if flow_input_names is None else flow_input_names
    flow_output_names = list() if flow_output_names is None else flow_output_names

    _validate_io_names_match(
        logic_flow=logic_flow,
        flow_input_names=flow_input_names,
        flow_output_names=flow_output_names,
    )

    graph = _create_flow_graph(
        logic_flow=logic_flow,
        flow_input_names=flow_input_names,
        flow_output_names=flow_output_names,
    )

    if not nx.algorithms.is_directed_acyclic_graph(graph):
        cycles = list(nx.algorithms.simple_cycles(graph))
        raise ValueError(CYCLE_ERROR_MSG + ". Cycles are: " + str(cycles))



from enum import Enum

from classiq_interface.generator.arith import binary_ops
from classiq_interface.generator.arith.register_user_input import RegisterUserInput


class CDFComparator(binary_ops.LessEqual):
    def create_io_enums(self):

        if isinstance(self.left_arg, RegisterUserInput) and isinstance(
            self.right_arg, RegisterUserInput
        ):
            self._input_enum = Enum(
                "BinaryOpInputs",
                {
                    self.left_arg.name: self.left_arg.name,
                    self.right_arg.name: self.right_arg.name,
                },
            )
            self._output_enum = Enum(
                "BinaryOpOutputs",
                {
                    self.output_name: self.output_name,
                    self.left_arg.name: self.left_arg.name,
                    self.right_arg.name: self.right_arg.name,
                },
            )
            return

        if isinstance(self.left_arg, RegisterUserInput):
            arg_name = self.left_arg.name
        else:
            assert isinstance(
                self.right_arg, RegisterUserInput
            ), "At least one argument should be a register"
            arg_name = self.right_arg.name

        self._input_enum = Enum("BinaryOpInputs", {arg_name: arg_name})
        self._output_enum = Enum(
            "BinaryOpOutputs",
            {self.output_name: self.output_name, arg_name: arg_name},
        )



import enum
from typing import Callable, Dict, List, Optional, Union

import pydantic


class WireRole(str, enum.Enum):
    INPUT = "input"
    OUTPUT = "output"
    AUXILIARY = "auxiliary"
    ZERO = "zero"


class Wire(pydantic.BaseModel):
    role: WireRole
    name: str
    qubit_indexes_relative: List[int]
    qubit_indexes_absolute: List[int] = list()

    def __len__(self):
        return self.qubit_indexes_relative.__len__()

    @property
    def width(self):
        return len(self)


class FunctionMetrics(pydantic.BaseModel):
    name: str
    wires: List[Wire] = list()
    depth: Optional[int]
    width: Optional[int]

    def __getitem__(self, key):
        if type(key) is int:
            return self.wires[key]
        if type(key) is str:
            for wire in self.wires:
                if key == wire.name:
                    return wire
        raise KeyError(key)

    def _update_wires(self, absolute_index_getter: Callable[[int], int]):
        for w in self.wires:
            w.qubit_indexes_absolute = list(
                map(absolute_index_getter, w.qubit_indexes_relative)
            )


class SynthesisMetrics(pydantic.BaseModel):
    duration_in_seconds: Optional[float]
    failure_reason: Optional[str]
    topological_sort: Optional[List[str]]
    function_metrics: List[FunctionMetrics] = pydantic.Field(default_factory=list)
    _function_mapping: Dict[
        Optional[Union[int, str]], FunctionMetrics
    ] = pydantic.PrivateAttr(default_factory=dict)

    def __getitem__(self, key) -> FunctionMetrics:
        if not self._function_mapping:
            for i, fm in enumerate(self.function_metrics):
                self._function_mapping[i] = fm
                self._function_mapping[fm.name] = fm

        try:
            return self._function_mapping[key]
        except KeyError:
            close_to_key = [
                map_key
                for map_key in self._function_mapping
                if isinstance(map_key, str) and map_key.startswith(key)
            ]
            if len(close_to_key) == 0:
                raise KeyError(f"No function named {key}")
            elif len(close_to_key) == 1:
                return self._function_mapping[close_to_key[0]]
            else:
                raise KeyError(f"Multiple function named {key} found, {close_to_key}")

    def __len__(self):
        return self.function_metrics.__len__()

    def __iter__(self):
        for function_name in self.topological_sort:
            yield self[function_name]

    def _update_wires(self, qubit_absolute_indexes: Dict[str, List[int]]):
        for fm in self.function_metrics:
            absolute_index_getter = qubit_absolute_indexes[fm.name].__getitem__
            fm._update_wires(absolute_index_getter)

    def pprint(self):
        print("Circuit Synthesis Metrics")
        print(f"    Generation took {self.duration_in_seconds} seconds")
        if self.failure_reason:
            print("Generation failed :(")
            print(f"Failure reason: {self.failure_reason}")
            return
        print(f"The circuit has {len(self.function_metrics)} functions:")
        for index, fm in enumerate(self.function_metrics):
            print(f"{index}) {fm.name}")
            if fm.name != "OUT":
                print(
                    f"  depth: {fm.depth} ; width: {fm.width} ; wires: {len(fm.wires)}"
                )
                for wire_index, wire in enumerate(fm.wires):
                    print(
                        f"  {wire_index}) {wire.role.value} - {wire.name} ; qubits: {wire.qubit_indexes_absolute}"
                    )



import abc
from enum import Enum
from typing import Any, List, Type

import pydantic

DEFAULT_OUTPUT_NAME = "OUT"
DEFAULT_INPUT_NAME = "IN"


class DefaultInputEnum(Enum):
    pass


class DefaultOutputEnum(Enum):
    OUT = DEFAULT_OUTPUT_NAME


class IO(Enum):
    Input = "Input"
    Output = "Output"


def input_io(is_inverse: bool) -> IO:
    if is_inverse:
        return IO.Output
    return IO.Input


def output_io(is_inverse: bool) -> IO:
    if is_inverse:
        return IO.Input
    return IO.Output


class FunctionParams(pydantic.BaseModel, abc.ABC):
    _input_enum: Type[Enum] = pydantic.PrivateAttr(default=None)
    _output_enum: Type[Enum] = pydantic.PrivateAttr(default=None)

    _input_names: List[str] = pydantic.PrivateAttr(default_factory=list)
    _output_names: List[str] = pydantic.PrivateAttr(default=[DEFAULT_OUTPUT_NAME])

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.create_io_enums()
        self._create_io_names()

    def get_io_enum(self, io: IO) -> Type[Enum]:
        if io == IO.Input:
            return self._input_enum
        elif io == IO.Output:
            return self._output_enum
        raise AssertionError("Unsupported IO type")

    def get_io_names(self, io: IO, is_inverse: bool = False) -> List[str]:
        assert io == IO.Input or io == IO.Output, "Unsupported IO type"
        if (io == IO.Input) ^ is_inverse:
            return self._input_names
        else:
            return self._output_names

    def create_io_enums(self) -> None:
        pass

    def _create_io_names(self) -> None:
        if self._input_enum is None and self._output_enum is None:
            return

        self._input_names = list(self._input_enum.__members__.keys())
        self._output_names = list(self._output_enum.__members__.keys())

    def is_valid_io_name(self, name: str, io: IO) -> bool:
        return name in self.get_io_names(io)



from typing import Optional

import pydantic


class RegisterUserInput(pydantic.BaseModel):
    size: pydantic.PositiveInt
    name: Optional[str] = None
    is_signed: bool = False
    fraction_places: pydantic.NonNegativeInt = 0

    class Config:
        extra = "forbid"



from typing import List

import pydantic

from classiq_interface.status import Status


class AnglesResult(pydantic.BaseModel):
    status: Status
    details: List[float]


class PyomoObjectResult(pydantic.BaseModel):
    status: Status
    details: str



from typing import List, Optional

import numpy as np
import pyomo.core as pyo


def knapsack(
    values: List[int],
    upper_bound: int,
    weights: Optional[List[int]] = None,
    max_weight: Optional[int] = None,
) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()

    model.x = pyo.Var(
        range(len(values)), domain=pyo.NonNegativeIntegers, bounds=(0, upper_bound)
    )

    if max_weight is not None and weights is not None:
        assert len(values) == len(
            weights
        ), "values and weights must be with the same length"
        model.weight_constraint = pyo.Constraint(
            expr=weights @ np.array(model.x.values()) <= max_weight
        )

    model.cost = pyo.Objective(
        expr=values @ np.array(model.x.values()), sense=pyo.maximize
    )

    return model
